import torch

from .base_modules import Linear


class TanhHead(torch.nn.Module):
    """Head for sentence-level classification tasks."""

    def __init__(self, input_dim: int, inner_dim: int, num_classes: int):
        super().__init__()
        self.dense = Linear(input_dim, inner_dim)
        self.out_proj = Linear(inner_dim, num_classes)

    def forward(self, hidden_states: torch.Tensor) -> torch.Tensor:
        hidden_states = self.dense(hidden_states)
        hidden_states = torch.tanh(hidden_states)
        return self.out_proj(hidden_states)



from typing import Union, List

import torch
from torch.nn.functional import relu


class FeedForward(torch.nn.Module):
    def __init__(self, input_dim: int, num_layers: int, dims: Union[int, List[int]]):

        super().__init__()
        if isinstance(dims, int):
            dims = [dims] * num_layers
        if len(dims) != num_layers:
            raise ValueError('len(hidden_dims) (%d) != num_layers (%d)' % (len(dims), num_layers))
        input_dims = [input_dim] + dims[:-1]
        self._linear_layers = torch.nn.ModuleList(
            torch.nn.Linear(layer_input_dim, layer_output_dim)
            for layer_input_dim, layer_output_dim in zip(input_dims, dims)
        )
        self._output_dim = dims[-1]
        self.input_dim = input_dim

    def get_output_dim(self):
        return self._output_dim

    def get_input_dim(self):
        return self.input_dim

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        for layer in self._linear_layers:
            tensor = relu(layer(tensor))
        return tensor



from typing import Any, Optional, Union, Tuple, List, Dict

import time
import select
import socket
import io
import threading
import operator
import uuid
import pickle as pkl
import torch

from ..utils import logger
from .socket_connection import SocketConnection
from .model_deployment_config import ModelDeploymentConfig
from .configs import Config


class ClientSocket(SocketConnection):
    __slots__ = ('message_id', 'batch_ind')

    def __init__(self, connection: socket.socket, message_size: int):
        super().__init__(connection, message_size)
        self.message_id: Optional[bytes] = None
        self.batch_ind: int = -1

    def request_server(self):
        self.connection.send(Config.request_signal)

    def cancel_request(self):
        self.connection.send(Config.request_cancelled)

    def send_inputs(self, data: Any, batch_ind: int):
        self.message_id = uuid.uuid4().bytes[:Config.message_id_len]
        buffer = io.BytesIO()
        torch.save(data, buffer)
        self.send_data(self.message_id + buffer.getvalue())
        self.batch_ind = batch_ind

    def receive_outputs(self) -> Optional[Tuple[int, Any]]:
        data = self.read_buffer()
        response_message_id = data[:Config.message_id_len]
        if response_message_id != self.message_id:
            logger.warning(f'Skipping response with wrong message id')
            return None
        data = data[Config.message_id_len:]
        self.message_id = None
        if data.startswith(Config.error_prefix):
            raise ConnectionError(data[len(Config.error_prefix):].decode('utf-8'))
        return self.batch_ind, pkl.loads(data)


class ClientAPI:
    client_sockets: Dict[bytes, Dict[str, List[ClientSocket]]] = {}

    def __init__(self, model: str, model_config: ModelDeploymentConfig):
        self.model = model
        self.num_models = len(model_config.allocations)
        self.port = model_config.port
        self.max_batch_size = model_config.max_batch_size

    @staticmethod
    def create_sockets(
            model: str, num_models: int, port: int, attempts: int, wait_time: float) -> List[ClientSocket]:
        logger.debug('Connecting to model workers')
        model_sockets: List[ClientSocket] = []
        for worker_ind in range(num_models):
            client_socket = socket.socket()
            for _ in range(attempts):
                try:
                    client_socket.connect((Config.host, port + worker_ind))
                    logger.debug(f'{model} model worker {worker_ind} connected')
                    model_sockets.append(
                        ClientSocket(connection=client_socket, message_size=Config.client_message_size))
                    break
                except ConnectionRefusedError:
                    time.sleep(wait_time)
        return model_sockets

    @staticmethod
    def split_inputs_to_batches(inputs: Dict, max_batch_size: int) -> List[Dict]:
        batch_size = next(iter(inputs.values())).shape[0]
        if max_batch_size is None or batch_size <= max_batch_size:
            return [inputs]
        return [
            {key: value[ind: ind + max_batch_size] for key, value in inputs.items()}
            for ind in range(0, batch_size, max_batch_size)
        ]

    @staticmethod
    def split_batch(inputs: Dict, batch_size: int) -> Dict:
        outputs = {}
        for key, values in inputs.items():
            outputs[key] = values[:batch_size]
            inputs[key] = values[batch_size:]
        return outputs

    @staticmethod
    def concat_lists_or_tensors(items: Union[List[List], List[torch.Tensor]]) -> Union[List, torch.Tensor]:
        return (
            torch.cat(items) if isinstance(items[0], torch.Tensor)
            else [value for sample_values in items for value in sample_values]
        )

    @classmethod
    def stack_outputs(cls, outputs: List) -> Any:
        sample = outputs[0]
        if len(outputs) == 1:
            return sample
        if isinstance(sample, dict):
            return {
                key: cls.concat_lists_or_tensors(list(map(operator.itemgetter(key), outputs))) for key in sample.keys()
            }
        return cls.concat_lists_or_tensors(outputs)

    @staticmethod
    def clear_sockets(socket_list: List[ClientSocket]):
        while True:
            active_sockets: List[ClientSocket] = select.select(socket_list, [], [], 0)[0]
            if not active_sockets:
                break
            for client_socket in active_sockets:
                client_socket.connection.recv(client_socket.message_size)

    def forward(self, **inputs):
        thread_id = threading.get_ident().to_bytes(length=Config.thread_id_len, byteorder='big', signed=False)

        if thread_id not in self.client_sockets:
            self.client_sockets[thread_id] = {}
        if self.model not in self.client_sockets[thread_id]:
            self.client_sockets[thread_id][self.model] = self.create_sockets(
                self.model, self.num_models, self.port, attempts=120, wait_time=10)

        client_sockets = self.client_sockets[thread_id][self.model]

        self.clear_sockets(client_sockets)

        total_batch_size = next(iter(inputs.values())).shape[0]

        outputs: List = []
        received_counter = 0
        cur_ind = 0

        for client_socket in client_sockets:
            client_socket.request_server()

        while total_batch_size or received_counter < len(outputs):
            ready_sockets: List[ClientSocket] = select.select(
                client_sockets if total_batch_size else
                [client_socket for client_socket in client_sockets if client_socket.message_id is not None],
                [], [], None
            )[0]
            input_sockets: List[ClientSocket] = []
            output_sockets: List[ClientSocket] = []
            for client_socket in ready_sockets:
                (input_sockets if client_socket.message_id is None else output_sockets).append(client_socket)

            added_connections = min(len(input_sockets), total_batch_size)
            outputs.extend(None for _ in range(added_connections))

            if added_connections:
                preferred_batch_size = (total_batch_size - 1) // added_connections + 1
                if self.max_batch_size is not None and preferred_batch_size > self.max_batch_size:
                    total_batch_size -= self.max_batch_size * added_connections
                    leftover_num = 0
                    batch_size = self.max_batch_size
                else:
                    leftover_num = total_batch_size % added_connections
                    total_batch_size = 0
                    batch_size = preferred_batch_size
                    if leftover_num:
                        batch_size -= 1
            else:
                batch_size = leftover_num = 0

            selected_sockets: List[ClientSocket] = input_sockets[:added_connections]

            self.clear_sockets(input_sockets)

            for ind, client_socket in enumerate(selected_sockets):
                sample_batch_size = batch_size
                if ind < leftover_num:
                    sample_batch_size += 1
                client_socket.send_inputs(self.split_batch(inputs, sample_batch_size), cur_ind + ind)

            cur_ind += added_connections

            if added_connections and not total_batch_size:
                for client_socket in client_sockets:
                    if client_socket.message_id is None:
                        client_socket.cancel_request()

            for client_socket in output_sockets:
                socket_outputs = client_socket.receive_outputs()
                if socket_outputs is None:
                    continue
                batch_ind, data = socket_outputs
                outputs[batch_ind] = data
                received_counter += 1
                if total_batch_size:
                    client_socket.request_server()

        return self.stack_outputs(outputs)

    def __call__(self, **kwargs):
        return self.forward(**kwargs)

    def close(self):
        for client_sockets in self.client_sockets.values():
            for sockets in client_sockets.values():
                for client_socket in sockets:
                    client_socket.close()



from typing import Optional, Tuple

import math

import torch
from torch.nn.functional import linear
import torch.jit

from ..modules import Linear, LayerNorm, Embedding, TransformerEncoder, TransformerDecoder
from ..utils import transposed_cat


class Blender(torch.nn.Module):
    def __init__(
            self, vocabulary_size: int, dim: int, n_positions: int, n_heads: int, ffn_size: int,
            n_encoder_layers: int, n_decoder_layers: int, num_labels: int, decode_step: int
    ):
        super().__init__()
        self.embeddings = Embedding(vocabulary_size, dim, 0)
        self.position_embeddings = Embedding(n_positions, dim)

        self.encoder_norm = LayerNorm(dim)
        self.encoder = self.build_encoder(
            n_heads=n_heads, n_encoder_layers=n_encoder_layers, dim=dim, ffn_size=ffn_size
        )

        self.pre_decoder_layer = Linear(dim, 2 * dim * n_decoder_layers)

        self.decoder_norm = LayerNorm(dim)
        self.decoder = self.build_decoder(
            n_heads=n_heads, n_decoder_layers=n_decoder_layers, embedding_size=dim, ffn_size=ffn_size
        )

        self.classifier_head = torch.nn.Linear(dim, num_labels) if num_labels else None
        self.scale = math.sqrt(dim)

        self.n_decoder_layers = n_decoder_layers
        self.n_heads = n_heads

        self.decode_step = decode_step

    @classmethod
    def from_pretrained(
            cls, model_path: str, vocabulary_size: int, embedding_size: int, n_positions: int, n_heads: int,
            ffn_size: int, n_encoder_layers: int, n_decoder_layers: int, num_labels: int, decode_step: int = 1
    ):
        model = cls(
            vocabulary_size=vocabulary_size, dim=embedding_size,
            n_positions=n_positions, n_heads=n_heads, ffn_size=ffn_size,
            n_encoder_layers=n_encoder_layers, n_decoder_layers=n_decoder_layers,
            num_labels=num_labels, decode_step=decode_step
        )
        model.load_state_dict(torch.load(model_path, map_location='cpu'))
        model = model

        model.eval()

        for param in model.parameters():
            param.requires_grad = False

        return model.half()

    def embedding_forward(self, tensor: torch.Tensor, positions: torch.Tensor) -> torch.Tensor:
        return self.embeddings(tensor) * self.scale + self.position_embeddings(positions)

    @torch.jit.export
    def encode(self, encoder_ids: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        mask = encoder_ids.to(torch.bool)
        positions = (mask.cumsum(dim=1, dtype=torch.long) - 1).clamp_(min=0)
        tensor = self.embedding_forward(encoder_ids, positions)
        tensor = self.encoder.forward(tensor, mask)
        tensor = self.encoder_norm(tensor)
        tensor = self.pre_decoder_layer(tensor).view(
            tensor.shape[:2] + (self.n_decoder_layers, 2, self.n_heads, -1)).transpose(1, 4).contiguous()
        return tensor, mask

    @torch.jit.export
    def decode(
            self, decoder_ids: torch.Tensor, encoder_state: torch.Tensor, encoder_mask: torch.Tensor,
            incr_state: Optional[torch.Tensor], get_incr_state: bool
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        mask = decoder_ids.to(torch.bool)

        positions = mask.cumsum(dim=1, dtype=torch.long)
        if incr_state is not None:
            incr_state = incr_state.flatten(0, 1)
            decoder_ids = decoder_ids[:, -self.decode_step:]
            positions = positions[:, -self.decode_step:]

        positions = (positions - 1).clamp_(min=0)

        tensor = self.embedding_forward(decoder_ids, positions)

        tensor, incr_state = self.decoder.forward(
            tensor=tensor, decoder_mask=mask, encoder_state=encoder_state, encoder_mask=encoder_mask,
            incr_state=incr_state, get_incr_state=get_incr_state
        )
        tensor = self.decoder_norm(tensor)
        return tensor, incr_state

    @classmethod
    def build_decoder(cls, n_heads, n_decoder_layers, embedding_size, ffn_size):
        return TransformerDecoder(
            n_heads=n_heads, n_layers=n_decoder_layers, embedding_size=embedding_size, ffn_size=ffn_size,
            blender_norm=True
        )

    @classmethod
    def build_encoder(cls, n_heads, n_encoder_layers, dim, ffn_size):
        return TransformerEncoder(
            n_heads=n_heads, n_layers=n_encoder_layers, embedding_size=dim, ffn_size=ffn_size, blender_norm=True
        )

    @classmethod
    def convert_old_state_dict(cls, state_dict: dict, n_encoder_layers: int, n_decoder_layers: int):
        for p in 'weight', 'bias':
            for module in 'encoder', 'decoder':
                state_dict[f'{module}_norm.{p}'] = state_dict.pop(f'{module}.norm_embeddings.{p}')
            state_dict[f'pre_decoder_layer.{p}'] = torch.cat(
                [state_dict.pop(f'decoder.layers.{ind}.encoder_attention.{name}_lin.{p}')
                 for ind in range(n_decoder_layers) for name in 'kv'], dim=0)
            for ind in range(n_encoder_layers):
                state_dict[f'encoder.layers.{ind}.attention.pre_attention.{p}'] = transposed_cat(
                    [state_dict.pop(f'encoder.layers.{ind}.attention.{name}_lin.{p}') for name in 'qkv'])
            for ind in range(n_decoder_layers):
                state_dict[f'decoder.layers.{ind}.self_attention.pre_attention.{p}'] = transposed_cat(
                    [state_dict.pop(f'decoder.layers.{ind}.self_attention.{name}_lin.{p}') for name in 'qkv'])

        state_dict['position_embeddings.weight'] = state_dict['encoder.position_embeddings.weight']

        keys_to_delete = [
            'criterion.weight', 'START',
            'encoder.embeddings.weight', 'encoder.position_embeddings.weight',
            'decoder.embeddings.weight', 'decoder.position_embeddings.weight'
        ]

        for key in keys_to_delete:
            state_dict.pop(key, None)

    def load_state_dict(self, state_dict: dict, strict=True):
        if 'model' in state_dict:
            state_dict = state_dict['model']

        if self.classifier_head is None:
            for p in ('weight', 'bias'):
                state_dict.pop(f'classifier_head.{p}', None)

        if 'encoder.pre_decoder_layer.weight' not in state_dict:
            self.convert_old_state_dict(
                state_dict, n_encoder_layers=len(self.encoder.layers), n_decoder_layers=len(self.decoder.layers))

        super().load_state_dict(state_dict, strict)

    @torch.jit.export
    def output(self, tensor: torch.Tensor) -> torch.Tensor:
        return linear(tensor, self.embeddings.weight)



import torch
from torch.nn import Module, ModuleList

from .base_modules import LayerNorm
from .attention import MultiHeadAttention
from .feed_forward import TransformerFFN


class TransformerEncoderLayer(Module):
    def __init__(self, n_heads: int, embedding_size: int, ffn_size: int, blender_norm: bool):
        super().__init__()
        self.attention = MultiHeadAttention(n_heads, embedding_size)
        self.norm1 = LayerNorm(embedding_size)

        self.ffn = TransformerFFN(embedding_size, ffn_size)
        self.norm2 = LayerNorm(embedding_size)

        self.blender_norm = blender_norm

    def forward(self, tensor: torch.Tensor, mask: torch.Tensor):
        residual = tensor
        if self.blender_norm:
            tensor = self.norm1(tensor)
        tensor = self.attention(tensor, mask=mask)[0]
        tensor = tensor + residual
        if not self.blender_norm:
            tensor = self.norm1(tensor)
        residual = tensor
        if self.blender_norm:
            tensor = self.norm2(tensor)
        tensor = self.ffn(tensor)
        tensor = tensor + residual
        if not self.blender_norm:
            tensor = self.norm2(tensor)
        return tensor


class TransformerEncoder(Module):
    def __init__(self, n_heads: int, n_layers: int, embedding_size: int, ffn_size: int, blender_norm: bool = False):
        super(TransformerEncoder, self).__init__()

        self.layers = ModuleList(
            TransformerEncoderLayer(n_heads, embedding_size, ffn_size, blender_norm) for _ in range(n_layers))
        self.n_heads = n_heads

    def forward(self, tensor: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
        attn_mask = ~mask.repeat_interleave(repeats=self.n_heads, dim=0).unsqueeze(1).expand(-1, mask.shape[1], -1)

        for layer in self.layers:
            tensor = layer.forward(tensor, attn_mask)

        return tensor * mask.unsqueeze(-1)



import os
import logging

import torch

from ..modules import Linear, TransformerEncoder, MaskEmbeddings
from ..utils import transposed_cat, neg_inf


logger = logging.getLogger(__name__)


class RobertaModel(torch.nn.Module):
    def __init__(
            self, vocab_size: int, dim: int, padding_idx: int, max_position_embeddings: int, ffn_size: int,
            n_heads: int, n_layers: int
    ):
        super().__init__()

        self.embeddings = MaskEmbeddings(
            vocab_size=vocab_size, dim=dim, padding_idx=padding_idx,
            max_position_embeddings=max_position_embeddings
        )
        self.encoder = TransformerEncoder(
            n_heads=n_heads, embedding_size=dim, ffn_size=ffn_size, n_layers=n_layers)

    def forward(self, input_ids: torch.Tensor, attention_mask: torch.Tensor) -> torch.Tensor:
        embedding_output = self.embeddings(input_ids=input_ids)
        return self.encoder.forward(embedding_output, attention_mask)


class RobertaForQuestionAnswering(torch.nn.Module):
    def __init__(
            self, vocab_size: int, dim: int, padding_idx: int,
            max_position_embeddings: int, ffn_size: int, n_heads: int, n_layers: int, num_labels: int):
        super().__init__()
        self.transformer = RobertaModel(
            vocab_size=vocab_size, dim=dim, padding_idx=padding_idx,
            max_position_embeddings=max_position_embeddings, ffn_size=ffn_size, n_heads=n_heads, n_layers=n_layers
        )
        self.qa_outputs = Linear(dim, num_labels)

    @classmethod
    def from_pretrained(cls, model_path: str) -> 'RobertaForQuestionAnswering':
        model = cls(
            vocab_size=50265, dim=768, padding_idx=1, max_position_embeddings=514,
            ffn_size=3072, n_heads=12, n_layers=12, num_labels=2
        )
        model.load_state_dict(state_dict=torch.load(os.path.join(model_path, 'pytorch_model.bin'), map_location='cpu'))
        model.eval()
        for p in model.parameters():
            p.requires_grad = False
        return model.half()

    def forward(self, input_ids: torch.Tensor, attention_mask: torch.Tensor, answer_mask: torch.Tensor) -> torch.Tensor:
        logits = self.transformer(input_ids, attention_mask=attention_mask)
        logits = self.qa_outputs(logits)
        logits.masked_fill_(~answer_mask.unsqueeze(-1), neg_inf(logits.dtype == torch.float16))
        logits = (logits - logits.exp().sum(dim=1, keepdim=True).log()).exp()
        logits[:, 0] = 0.
        return logits

    def load_state_dict(self, state_dict, strict=True):
        module_names = {'layers': 'layer', 'transformer': 'roberta'}
        for key, value in list(state_dict.items()):
            new_key = key
            for name, old_name in module_names.items():
                new_key = new_key.replace(f'{old_name}.', f'{name}.')
            state_dict[new_key] = state_dict.pop(key)
        del state_dict['transformer.embeddings.position_ids']
        old_names = {
            'attention.out_lin': 'attention.output.dense',
            'norm1': 'attention.output.LayerNorm',
            'norm2': 'output.LayerNorm',
            'ffn.lin1': 'intermediate.dense',
            'ffn.lin2': 'output.dense'
        }
        state_dict['transformer.embeddings.word_embeddings.weight'] += state_dict.pop(
            'transformer.embeddings.token_type_embeddings.weight')
        for layer_ind in range(len(self.transformer.encoder.layers)):
            prefix = f'transformer.encoder.layers.{layer_ind}.'
            for p in ('weight', 'bias'):
                state_dict[f'{prefix}attention.pre_attention.{p}'] = transposed_cat([
                    state_dict.pop(f'{prefix}attention.self.{k}.{p}') for k in ('query', 'key', 'value')
                ])
                for new_name, old_name in old_names.items():
                    state_dict[f'{prefix}{new_name}.{p}'] = state_dict.pop(f'{prefix}{old_name}.{p}')
        return super().load_state_dict(state_dict, strict)



from typing import Any, Tuple, Dict

import copy
import os
import json
import math

import torch
from torch import nn
from torch.nn.functional import relu, softmax

from transformers.modeling_outputs import BaseModelOutput, BaseModelOutputWithPastAndCrossAttentions, Seq2SeqLMOutput
from transformers.generation_utils import GenerationMixin


class Linear(torch.nn.Linear):
    def __init__(self, in_features: int, out_features: int, bias: bool = True):
        torch.nn.Module.__init__(self)
        self.in_features = in_features
        self.out_features = out_features
        self.weight = nn.Parameter(torch.Tensor(out_features, in_features))
        if bias:
            self.bias = nn.Parameter(torch.Tensor(out_features))
        else:
            self.register_parameter('bias', None)


class T5LayerNorm(nn.Module):
    def __init__(self, hidden_size, eps=1e-6):
        """
        Construct a layernorm module in the T5 style No bias and no subtraction of mean.
        """
        super().__init__()
        self.weight = nn.Parameter(torch.ones(hidden_size))
        self.variance_epsilon = eps

    def forward(self, hidden_states):
        # layer norm should always be calculated in float32
        variance = hidden_states.to(torch.float32).pow(2).mean(-1, keepdim=True)
        hidden_states = hidden_states * torch.rsqrt(variance + self.variance_epsilon)
        return self.weight * hidden_states.to(self.weight.dtype)


class T5DenseReluDense(nn.Module):
    def __init__(self, d_model: int, d_ff: int):
        super().__init__()
        self.wi = Linear(d_model, d_ff, bias=False)
        self.wo = Linear(d_ff, d_model, bias=False)

    def forward(self, hidden_states: torch.Tensor) -> torch.Tensor:
        hidden_states = self.wi(hidden_states)
        hidden_states = relu(hidden_states)
        return self.wo(hidden_states)


class T5LayerFF(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.DenseReluDense = T5DenseReluDense(d_model=config['d_model'], d_ff=config['d_ff'])
        self.layer_norm = T5LayerNorm(config['d_model'], eps=config['layer_norm_epsilon'])

    def forward(self, hidden_states: torch.Tensor) -> torch.Tensor:
        forwarded_states = self.layer_norm(hidden_states)
        forwarded_states = self.DenseReluDense(forwarded_states)
        return hidden_states + forwarded_states


class T5Attention(nn.Module):
    def __init__(self, config: dict, has_relative_attention_bias=False):
        super().__init__()
        self.is_decoder = config['is_decoder']
        self.has_relative_attention_bias = has_relative_attention_bias

        self.relative_attention_num_buckets = config['relative_attention_num_buckets']
        self.d_model = config['d_model']
        self.key_value_proj_dim = config['d_kv']
        self.n_heads = config['num_heads']
        self.inner_dim = self.n_heads * self.key_value_proj_dim

        # Mesh TensorFlow initialization to avoid scaling before softmax
        self.q = Linear(self.d_model, self.inner_dim, bias=False)
        self.k = Linear(self.d_model, self.inner_dim, bias=False)
        self.v = Linear(self.d_model, self.inner_dim, bias=False)
        self.o = Linear(self.inner_dim, self.d_model, bias=False)

        if self.has_relative_attention_bias:
            self.relative_attention_bias = nn.Embedding(self.relative_attention_num_buckets, self.n_heads)

    @staticmethod
    def _relative_position_bucket(
            relative_position, bidirectional=True, num_buckets=32, max_distance=128) -> torch.Tensor:
        if bidirectional:
            num_buckets //= 2
            relative_buckets = (relative_position > 0).to(torch.long) * num_buckets
            relative_position = torch.abs(relative_position)
        else:
            relative_buckets = 0
            relative_position = -torch.min(relative_position, torch.zeros_like(relative_position))
        # now relative_position is in the range [0, inf)

        # half of the buckets are for exact increments in positions
        max_exact = num_buckets // 2
        is_small = relative_position < max_exact

        # The other half of the buckets are for logarithmically bigger bins in positions up to max_distance
        relative_postion_if_large = max_exact + (
            torch.log(relative_position.float() / max_exact)
            / math.log(max_distance / max_exact)
            * (num_buckets - max_exact)
        ).to(torch.long)
        relative_postion_if_large = torch.min(
            relative_postion_if_large, torch.full_like(relative_postion_if_large, num_buckets - 1)
        )

        return torch.where(is_small, relative_position, relative_postion_if_large) + relative_buckets

    def compute_bias(self, query_length: int, key_length: int) -> torch.Tensor:
        """ Compute binned relative position bias """
        context_position = torch.arange(query_length, dtype=torch.long).unsqueeze(1)
        memory_position = torch.arange(key_length, dtype=torch.long).unsqueeze(0)
        relative_position = memory_position - context_position  # shape (query_length, key_length)
        relative_position_bucket = self._relative_position_bucket(
            relative_position,  # shape (query_length, key_length)
            bidirectional=(not self.is_decoder),
            num_buckets=self.relative_attention_num_buckets,
        )
        relative_position_bucket = relative_position_bucket.to(self.relative_attention_bias.weight.device)
        values = self.relative_attention_bias(relative_position_bucket)  # shape (query_length, key_length, num_heads)
        return values.permute([2, 0, 1]).unsqueeze(0)  # shape (1, num_heads, query_length, key_length)

    def forward(
        self,
        hidden_states,
        mask=None,
        key_value_states=None,
        position_bias=None,
        past_key_value=None,
        layer_head_mask=None,
        query_length=None,
        use_cache=False
    ):
        """
        Self-attention (if key_value_states is None) or attention over source sentence (provided by key_value_states).
        """
        # Input is (batch_size, seq_length, dim)
        # Mask is (batch_size, key_length) (non-causal) or (batch_size, key_length, key_length)
        # past_key_value[0] is (batch_size, n_heads, q_len - 1, dim_per_head)
        batch_size, seq_length = hidden_states.shape[:2]

        real_seq_length = seq_length

        if past_key_value is not None:
            real_seq_length += past_key_value[0].shape[2] if query_length is None else query_length

        key_length = real_seq_length if key_value_states is None else key_value_states.shape[1]

        def shape(states):
            """  projection """
            return states.view(batch_size, -1, self.n_heads, self.key_value_proj_dim).transpose(1, 2)

        def unshape(states):
            """  reshape """
            return states.transpose(1, 2).contiguous().view(batch_size, -1, self.inner_dim)

        def project(hidden_states, proj_layer, key_value_states, past_key_value):
            """ projects hidden states correctly to key/query states """
            if key_value_states is None:
                # self-attn
                # (batch_size, n_heads, seq_length, dim_per_head)
                hidden_states = shape(proj_layer(hidden_states))
            elif past_key_value is None:
                # cross-attn
                # (batch_size, n_heads, seq_length, dim_per_head)
                hidden_states = shape(proj_layer(key_value_states))

            if past_key_value is not None:
                if key_value_states is None:
                    # self-attn
                    # (batch_size, n_heads, key_length, dim_per_head)
                    hidden_states = torch.cat([past_key_value, hidden_states], dim=2)
                else:
                    # cross-attn
                    hidden_states = past_key_value
            return hidden_states

        # get query states
        query_states = shape(self.q(hidden_states))  # (batch_size, n_heads, seq_length, dim_per_head)

        # get key/value states
        key_states = project(
            hidden_states, self.k, key_value_states, past_key_value[0] if past_key_value is not None else None
        )
        value_states = project(
            hidden_states, self.v, key_value_states, past_key_value[1] if past_key_value is not None else None
        )

        # compute scores
        scores = torch.matmul(query_states, key_states.transpose(3, 2))

        if position_bias is None:
            position_bias = (
                self.compute_bias(real_seq_length, key_length) if self.has_relative_attention_bias else
                torch.zeros((1, self.n_heads, real_seq_length, key_length), device=scores.device, dtype=scores.dtype)
            )

            # if key and values are already calculated
            # we want only the last query position bias
            if past_key_value is not None:
                position_bias = position_bias[:, :, -seq_length:, :]

            if mask is not None:
                position_bias = position_bias + mask  # (batch_size, n_heads, seq_length, key_length)

        scores += position_bias
        attn_weights = softmax(scores.float(), dim=-1).type_as(scores)  # (batch_size, n_heads, seq_length, key_length)

        # Mask heads if we want to
        if layer_head_mask is not None:
            attn_weights = attn_weights * layer_head_mask

        attn_output = unshape(torch.matmul(attn_weights, value_states))  # (batch_size, seq_length, dim)
        attn_output = self.o(attn_output)

        present_key_value_state = (key_states, value_states) if (self.is_decoder and use_cache) else None
        return attn_output, present_key_value_state, position_bias


class T5LayerSelfAttention(nn.Module):
    def __init__(self, config: dict, has_relative_attention_bias=False):
        super().__init__()
        self.SelfAttention = T5Attention(config, has_relative_attention_bias=has_relative_attention_bias)
        self.layer_norm = T5LayerNorm(config['d_model'], eps=config['layer_norm_epsilon'])

    def forward(
        self,
        hidden_states,
        attention_mask=None,
        position_bias=None,
        layer_head_mask=None,
        past_key_value=None,
        use_cache=False
    ):
        normed_hidden_states = self.layer_norm(hidden_states)
        attention_output = self.SelfAttention(
            normed_hidden_states,
            mask=attention_mask,
            position_bias=position_bias,
            layer_head_mask=layer_head_mask,
            past_key_value=past_key_value,
            use_cache=use_cache
        )
        return (hidden_states + attention_output[0],) + attention_output[1:]  # add attentions if we output them


class T5LayerCrossAttention(nn.Module):
    def __init__(self, config: dict):
        super().__init__()
        self.EncDecAttention = T5Attention(config, has_relative_attention_bias=False)
        self.layer_norm = T5LayerNorm(config['d_model'], eps=config['layer_norm_epsilon'])

    def forward(
        self,
        hidden_states,
        key_value_states,
        attention_mask=None,
        position_bias=None,
        layer_head_mask=None,
        past_key_value=None,
        use_cache=False,
        query_length=None
    ):
        normed_hidden_states = self.layer_norm(hidden_states)
        attention_output = self.EncDecAttention(
            normed_hidden_states,
            mask=attention_mask,
            key_value_states=key_value_states,
            position_bias=position_bias,
            layer_head_mask=layer_head_mask,
            past_key_value=past_key_value,
            use_cache=use_cache,
            query_length=query_length
        )
        return (hidden_states + attention_output[0],) + attention_output[1:]


class T5Block(nn.Module):
    def __init__(self, config: dict, has_relative_attention_bias=False):
        super().__init__()
        self.is_decoder = config['is_decoder']
        self.layer = nn.ModuleList()
        self.layer.append(T5LayerSelfAttention(config, has_relative_attention_bias=has_relative_attention_bias))
        if self.is_decoder:
            self.layer.append(T5LayerCrossAttention(config))

        self.layer.append(T5LayerFF(config))

    def forward(
        self,
        hidden_states,
        attention_mask=None,
        position_bias=None,
        encoder_hidden_states=None,
        encoder_attention_mask=None,
        encoder_decoder_position_bias=None,
        layer_head_mask=None,
        encoder_layer_head_mask=None,
        past_key_value=None,
        use_cache=False
    ):

        if past_key_value is not None:
            self_attn_past_key_value = past_key_value[:2]
            cross_attn_past_key_value = past_key_value[2:]
        else:
            self_attn_past_key_value, cross_attn_past_key_value = None, None

        self_attention_outputs = self.layer[0](
            hidden_states,
            attention_mask=attention_mask,
            position_bias=position_bias,
            layer_head_mask=layer_head_mask,
            past_key_value=self_attn_past_key_value,
            use_cache=use_cache
        )
        hidden_states, present_key_value_state = self_attention_outputs[:2]
        attention_outputs = self_attention_outputs[2:]  # Keep self-attention outputs and relative position weights

        # clamp inf values to enable fp16 training
        if torch.isinf(hidden_states).any():
            clamp_value = torch.finfo(hidden_states.dtype).max - 1000
            hidden_states = torch.clamp(hidden_states, min=-clamp_value, max=clamp_value)

        do_cross_attention = self.is_decoder and encoder_hidden_states is not None
        if do_cross_attention:
            # the actual query length is unknown for cross attention
            # if using past key value states. Need to inject it here
            query_length = None if present_key_value_state is None else present_key_value_state[0].shape[2]

            cross_attention_outputs = self.layer[1](
                hidden_states,
                key_value_states=encoder_hidden_states,
                attention_mask=encoder_attention_mask,
                position_bias=encoder_decoder_position_bias,
                layer_head_mask=encoder_layer_head_mask,
                past_key_value=cross_attn_past_key_value,
                query_length=query_length,
                use_cache=use_cache
            )
            hidden_states = cross_attention_outputs[0]
            if torch.isinf(hidden_states).any():
                clamp_value = torch.finfo(hidden_states.dtype).max - 1000
                hidden_states = torch.clamp(hidden_states, min=-clamp_value, max=clamp_value)

            # Combine self attn and cross attn key value states
            if present_key_value_state is not None:
                present_key_value_state = present_key_value_state + cross_attention_outputs[1]

            # Keep cross-attention outputs and relative position weights
            attention_outputs = attention_outputs + cross_attention_outputs[2:]

        # Apply Feed Forward layer
        hidden_states = self.layer[-1](hidden_states)
        if torch.isinf(hidden_states).any():
            clamp_value = torch.finfo(hidden_states.dtype).max - 1000
            hidden_states = torch.clamp(hidden_states, min=-clamp_value, max=clamp_value)

        return (hidden_states, present_key_value_state) + attention_outputs


class T5Stack(torch.nn.Module):
    def __init__(self, config: dict, embed_tokens=None):
        super().__init__()

        self.embed_tokens = embed_tokens
        self.is_decoder = config['is_decoder']

        self.block = nn.ModuleList(
            T5Block(config, has_relative_attention_bias=(i == 0)) for i in range(config['num_layers'])
        )
        self.final_layer_norm = T5LayerNorm(config['d_model'], eps=config['layer_norm_epsilon'])
        self.use_cache = config['use_cache']
        self.use_return_dict = True
        self.num_layers = config['num_layers']

    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        encoder_hidden_states=None,
        encoder_attention_mask=None,
        past_key_values=None,
        use_cache=None,
        return_dict=None
    ):
        use_cache = use_cache if use_cache is not None else self.use_cache
        return_dict = return_dict if return_dict is not None else self.use_return_dict

        input_shape = input_ids.size()
        input_ids = input_ids.view(-1, input_shape[-1])
        hidden_states = self.embed_tokens(input_ids)

        batch_size, seq_length = input_shape

        # required mask seq length can be calculated via length of past
        mask_seq_length = past_key_values[0][0].shape[2] + seq_length if past_key_values is not None else seq_length

        attention_mask = torch.ones(batch_size, mask_seq_length).to(hidden_states.device)
        if self.is_decoder and encoder_attention_mask is None and encoder_hidden_states is not None:
            encoder_seq_length = encoder_hidden_states.shape[1]
            encoder_attention_mask = torch.ones(
                batch_size, encoder_seq_length, device=hidden_states.device, dtype=torch.long
            )

        # initialize past_key_values with `None` if past does not exist
        if past_key_values is None:
            past_key_values = [None] * len(self.block)

        # ourselves in which case we just need to make it broadcastable to all heads.
        extended_attention_mask = self.get_extended_attention_mask(attention_mask, input_shape, hidden_states.device)

        encoder_extended_attention_mask = (
            self.invert_attention_mask(encoder_attention_mask)
            if self.is_decoder and encoder_attention_mask is not None else None
        )

        present_key_value_states = () if use_cache else None
        position_bias = None
        encoder_decoder_position_bias = None

        for layer_module, past_key_value in zip(self.block, past_key_values):
            layer_outputs = layer_module(
                hidden_states,
                attention_mask=extended_attention_mask,
                position_bias=position_bias,
                encoder_hidden_states=encoder_hidden_states,
                encoder_attention_mask=encoder_extended_attention_mask,
                encoder_decoder_position_bias=encoder_decoder_position_bias,
                layer_head_mask=None,
                encoder_layer_head_mask=None,
                past_key_value=past_key_value,
                use_cache=use_cache
            )
            # layer_outputs is a tuple with:
            # hidden-states, key-value-states, (self-attention weights), (self-attention position bias),
            # (cross-attention weights), (cross-attention position bias)
            hidden_states, present_key_value_state, position_bias = layer_outputs[:3]

            # We share the position biases between the layers - the first layer store them
            # layer_outputs = hidden-states, key-value-states (self-attention weights),
            # (self-attention position bias), (cross-attention weights), (cross-attention position bias)
            if self.is_decoder and encoder_hidden_states is not None:
                encoder_decoder_position_bias = layer_outputs[3]
            # append next layer key value states
            if use_cache:
                present_key_value_states = present_key_value_states + (present_key_value_state,)

        hidden_states = self.final_layer_norm(hidden_states)

        if not return_dict:
            return tuple(v for v in (hidden_states, present_key_value_states) if v is not None)
        return BaseModelOutputWithPastAndCrossAttentions(
            last_hidden_state=hidden_states, past_key_values=present_key_value_states
        )

    def get_extended_attention_mask(
            self, attention_mask: torch.Tensor, input_shape: Tuple[int, int], device: torch.device) -> torch.Tensor:
        # We can provide a self-attention mask of dimensions [batch_size, from_seq_length, to_seq_length]
        # ourselves in which case we just need to make it broadcastable to all heads.
        extended_attention_mask = attention_mask.unsqueeze(1)
        if attention_mask.dim() == 2:
            # Provided a padding mask of dimensions [batch_size, seq_length]
            # - if the model is a decoder, apply a causal mask in addition to the padding mask
            # - if the model is an encoder,
            # make the mask broadcastable to [batch_size, num_heads, seq_length, seq_length]
            extended_attention_mask = extended_attention_mask.unsqueeze(1)
            if self.is_decoder:
                batch_size, seq_length = input_shape
                seq_ids = torch.arange(seq_length, device=device).unsqueeze(0)
                causal_mask: torch.Tensor = seq_ids.unsqueeze(0).repeat(batch_size, seq_length, 1) <= seq_ids.unsqueeze(-1)
                # in case past_key_values are used we need to add a prefix ones mask to the causal mask
                # causal and attention masks must have same type with pytorch version < 1.3
                causal_mask = causal_mask.to(attention_mask.dtype)

                if causal_mask.shape[1] < attention_mask.shape[1]:
                    prefix_seq_len = attention_mask.shape[1] - causal_mask.shape[1]
                    causal_mask = torch.cat([
                        torch.ones((batch_size, seq_length, prefix_seq_len), device=device, dtype=causal_mask.dtype),
                        causal_mask
                    ], axis=-1)

                extended_attention_mask = causal_mask.unsqueeze(1) * extended_attention_mask

        # Since attention_mask is 1.0 for positions we want to attend and 0.0 for
        # masked positions, this operation will create a tensor which is 0.0 for
        # positions we want to attend and -10000.0 for masked positions.
        # Since we are adding it to the raw scores before the softmax, this is
        # effectively the same as removing these entirely.
        return (1.0 - extended_attention_mask.to(dtype=next(self.parameters()).dtype)) * -10000.0

    def invert_attention_mask(self, encoder_attention_mask: torch.Tensor) -> torch.Tensor:
        encoder_extended_attention_mask = encoder_attention_mask.unsqueeze(1)

        if encoder_attention_mask.dim() == 2:
            encoder_extended_attention_mask = encoder_extended_attention_mask.unsqueeze(1)

        dtype = next(self.parameters()).dtype

        return (1 - encoder_extended_attention_mask.to(dtype=dtype)) * (-1e4 if dtype == torch.float16 else -1e9)


class ConfigNameSpace:
    def __init__(self, kwargs: dict):
        self.max_length = 20
        self.torchscript = kwargs.pop("torchscript", False)  # Only used by PyTorch models
        self.use_bfloat16 = kwargs.pop("use_bfloat16", False)
        self.tie_word_embeddings = kwargs.pop("tie_word_embeddings", True)

        # Is decoder is used in encoder-decoder models to differentiate encoder from decoder
        self.is_encoder_decoder = kwargs.pop("is_encoder_decoder", True)
        self.is_decoder = kwargs.pop("is_decoder", False)
        self.add_cross_attention = kwargs.pop("add_cross_attention", False)
        self.tie_encoder_decoder = kwargs.pop("tie_encoder_decoder", False)

        # Parameters for sequence generation
        self.max_length = kwargs.pop("max_length", 20)
        self.min_length = kwargs.pop("min_length", 0)
        self.do_sample = kwargs.pop("do_sample", False)
        self.early_stopping = kwargs.pop("early_stopping", False)
        self.num_beams = kwargs.pop("num_beams", 1)
        self.num_beam_groups = kwargs.pop("num_beam_groups", 1)
        self.diversity_penalty = kwargs.pop("diversity_penalty", 0.0)
        self.temperature = kwargs.pop("temperature", 1.0)
        self.top_k = kwargs.pop("top_k", 50)
        self.top_p = kwargs.pop("top_p", 1.0)
        self.repetition_penalty = kwargs.pop("repetition_penalty", 1.0)
        self.length_penalty = kwargs.pop("length_penalty", 1.0)
        self.no_repeat_ngram_size = kwargs.pop("no_repeat_ngram_size", 0)
        self.encoder_no_repeat_ngram_size = kwargs.pop("encoder_no_repeat_ngram_size", 0)
        self.bad_words_ids = kwargs.pop("bad_words_ids", None)
        self.num_return_sequences = kwargs.pop("num_return_sequences", 1)
        self.chunk_size_feed_forward = kwargs.pop("chunk_size_feed_forward", 0)
        self.output_scores = kwargs.pop("output_scores", False)
        self.return_dict_in_generate = kwargs.pop("return_dict_in_generate", False)

        self.pad_token_id = 0
        self.eos_token_id = 1
        self.bos_token_id = None
        self.use_cache = kwargs.pop("use_cache", True)
        self.decoder_start_token_id = kwargs.pop('decoder_start_token_id')

        self.forced_bos_token_id = None
        self.forced_eos_token_id = None
        self.remove_invalid_values = None

        self.output_attentions = None
        self.output_hidden_states = None


class T5ForConditionalGeneration(torch.nn.Module, GenerationMixin):
    def __init__(self, config: dict):
        super().__init__()
        self.config = ConfigNameSpace(config)
        self.model_dim = config['d_model']

        self.shared = nn.Embedding(config['vocab_size'], config['d_model'])

        encoder_config = copy.deepcopy(config)
        encoder_config['is_decoder'] = False
        encoder_config['use_cache'] = False
        encoder_config['is_encoder_decoder'] = False
        self.encoder = T5Stack(encoder_config, self.shared)

        decoder_config = copy.deepcopy(config)
        decoder_config['is_decoder'] = True
        decoder_config['is_encoder_decoder'] = False
        decoder_config['num_layers'] = config['num_decoder_layers']
        decoder_config['use_cache'] = True

        self.decoder = T5Stack(decoder_config, self.shared)
        self.lm_head = Linear(config['d_model'], config['vocab_size'], bias=False)

        self.num_layers = config['num_layers']
        self.num_decoder_layers = config['num_decoder_layers']
        self.use_cache = True
        self.use_return_dict = True
        self.tie_word_embeddings = True

    @classmethod
    def from_pretrained(cls, model_path: str):
        with open(os.path.join(model_path, 'config.json'), 'r') as file:
            config = json.load(file)
        model = cls(config)
        state_dict = torch.load(os.path.join(model_path, 'pytorch_model.bin'), map_location='cpu')
        state_dict.pop('decoder.block.0.layer.1.EncDecAttention.relative_attention_bias.weight', None)
        model.load_state_dict(state_dict)
        return model

    def forward(
        self,
        attention_mask=None,
        decoder_input_ids=None,
        decoder_attention_mask=None,
        encoder_outputs=None,
        past_key_values=None,
        return_dict=None,
        output_hidden_states=None,
        output_attentions=None
    ):
        if return_dict is None:
            return_dict = self.use_return_dict

        # Encode if needed (training, first prediction pass)
        if return_dict and not isinstance(encoder_outputs, BaseModelOutput):
            encoder_outputs = BaseModelOutput(last_hidden_state=encoder_outputs[0])

        hidden_states = encoder_outputs[0]

        # If decoding with past key value states, only the last tokens
        # should be given as an input
        if past_key_values is not None:
            if decoder_input_ids is not None:
                decoder_input_ids = decoder_input_ids[:, -1:]

        # Decode
        decoder_outputs = self.decoder(
            input_ids=decoder_input_ids,
            attention_mask=decoder_attention_mask,
            past_key_values=past_key_values,
            encoder_hidden_states=hidden_states,
            encoder_attention_mask=attention_mask,
            use_cache=True,
            return_dict=return_dict
        )

        sequence_output = decoder_outputs[0]

        if self.tie_word_embeddings:
            sequence_output = sequence_output * (self.model_dim ** -0.5)

        lm_logits = self.lm_head(sequence_output)

        if not return_dict:
            return (lm_logits,) + decoder_outputs[1:] + encoder_outputs

        return Seq2SeqLMOutput(
            logits=lm_logits,
            past_key_values=decoder_outputs.past_key_values,
            encoder_last_hidden_state=encoder_outputs.last_hidden_state
        )

    @staticmethod
    def prepare_inputs_for_generation(
        input_ids, past=None, attention_mask=None, use_cache=None, encoder_outputs=None, **kwargs
    ):

        # cut decoder_input_ids if past is used
        if past is not None:
            input_ids = input_ids[:, -1:]

        return {
            "decoder_input_ids": input_ids,
            "past_key_values": past,
            "encoder_outputs": encoder_outputs,
            "attention_mask": attention_mask
        }

    def _prepare_encoder_decoder_kwargs_for_generation(self, input_ids: torch.Tensor, model_kwargs) -> Dict[str, Any]:
        model_kwargs["encoder_outputs"] = self.encoder(input_ids, return_dict=True)
        return model_kwargs



from typing import Optional, Tuple

import math

import torch
from torch.nn import Module

from .base_modules import Linear
from ..utils import neg_inf


class MultiHeadAttention(Module):
    def __init__(self, n_heads: int, dim: int):
        super().__init__()
        self.pre_attention = Linear(dim, dim * 3)
        self.out_lin = Linear(dim, dim)

        self.n_heads = n_heads
        self.dim_per_head = dim // n_heads
        self.scale = math.sqrt(self.dim_per_head)

    def forward(
            self, query: torch.Tensor, mask: torch.Tensor,
            incr_state: Optional[torch.Tensor] = None, get_incr_state: bool = False
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        batch_size, seq_len, dim = query.shape

        query = self.pre_attention(query)
        query = query.view(batch_size, seq_len, self.n_heads, self.dim_per_head, 3)
        query = query.transpose(1, 2).contiguous()
        query = query.view(batch_size * self.n_heads, seq_len, self.dim_per_head, 3)

        q = query[:, :, :, 0] / self.scale
        k_v = query[:, :, :, 1:]

        if incr_state is not None:
            k_v = torch.cat([incr_state, k_v], dim=1)

        incr_state = k_v.view(batch_size, self.n_heads, -1, self.dim_per_head, 2) if get_incr_state else None

        dot_prod = q.bmm(k_v[:, :, :, 0].transpose(1, 2))

        dot_prod.masked_fill_(mask, neg_inf(dot_prod.dtype == torch.float16))

        attn_weights = dot_prod.softmax(dim=-1, dtype=torch.float).type_as(query)

        return self.out_lin(attn_weights.bmm(k_v[:, :, :, 1]).view(
            batch_size, self.n_heads, seq_len,
            self.dim_per_head).transpose(1, 2).contiguous().view(batch_size, seq_len, dim)), incr_state


class DecoderEncoderAttention(Module):
    def __init__(self, n_heads: int, dim: int):
        super().__init__()
        self.q_lin = Linear(dim, dim)
        self.out_lin = Linear(dim, dim)

        self.n_heads = n_heads
        self.dim_per_head = dim // n_heads
        self.scale = math.sqrt(self.dim_per_head)

    def forward(self, query: torch.Tensor, mask: torch.Tensor, key: torch.Tensor, value: torch.Tensor) -> torch.Tensor:
        batch_size, seq_len, dim = query.shape

        query = self.q_lin(query) / self.scale
        query = query.view(batch_size, seq_len, self.n_heads, self.dim_per_head).transpose(1, 2).contiguous()
        query = query.view(batch_size * self.n_heads, seq_len, self.dim_per_head)

        dot_prod = query.bmm(key.transpose(1, 2))

        dot_prod.masked_fill_(mask, neg_inf(dot_prod.dtype == torch.float16))

        attn_weights = dot_prod.softmax(dim=-1, dtype=torch.float).type_as(query)

        return self.out_lin(attn_weights.bmm(value).view(
            batch_size, self.n_heads, seq_len,
            self.dim_per_head).transpose(1, 2).contiguous().view(batch_size, seq_len, dim))



from .base_modules import Linear, Embedding, LayerNorm
from .heads import TanhHead
from .embeddings import Embeddings, MaskEmbeddings, EmbeddingsWithTokenTypes
from .transformer_encoder import TransformerEncoder
from .transformer_decoder import TransformerDecoder



from .bart import BartForSequenceClassification
from .blender import Blender
from .distilbert import DistilBertForMaskedLM
from .electra import ElectraForSequenceClassification
from .roberta import RobertaForQuestionAnswering



import os

import torch
from torch import Tensor
from torch.nn import Module

from ..modules import Linear, LayerNorm, Embedding, TransformerEncoder, TransformerDecoder, TanhHead
from ..utils import transposed_cat


class BartModel(Module):
    def __init__(
            self, vocab_size: int, padding_idx: int, eos_token_id: int, dim: int, max_position_embeddings: int,
            encoder_layers: int, encoder_heads: int, encoder_ffn_dim: int,
            decoder_layers: int, decoder_heads: int, decoder_ffn_dim: int
    ):
        super().__init__()

        self.embeddings = Embedding(vocab_size, dim, padding_idx)

        self.encoder_positions = Embedding(
            num_embeddings=max_position_embeddings, embedding_dim=dim, padding_idx=padding_idx)

        self.encoder_norm = LayerNorm(dim)

        self.encoder = TransformerEncoder(
            n_heads=encoder_heads, n_layers=encoder_layers, embedding_size=dim, ffn_size=encoder_ffn_dim)

        self.pre_decoder_layer = Linear(dim, 2 * dim * decoder_layers)

        self.decoder_positions = Embedding(
            num_embeddings=max_position_embeddings, embedding_dim=dim, padding_idx=padding_idx)

        self.decoder_norm = LayerNorm(dim)

        self.encoder_heads = encoder_heads
        self.decoder_layers = decoder_layers

        self.decoder = TransformerDecoder(
            n_heads=decoder_heads, n_layers=decoder_layers, embedding_size=dim, ffn_size=decoder_ffn_dim)

        self.eos_token_id = eos_token_id

    def forward(self, input_ids: Tensor, attention_mask: Tensor):
        attention_mask = attention_mask.to(torch.bool)
        decoder_input_ids = torch.cat(
            [torch.empty((input_ids.shape[0], 1), dtype=torch.long, device=input_ids.device).fill_(self.eos_token_id),
             input_ids[:, :-1]], dim=1)
        decoder_padding_mask = torch.cat(
            [torch.ones((input_ids.shape[0], 1), dtype=torch.bool, device=input_ids.device), attention_mask[:, :-1]],
            dim=1)

        x = self.embeddings(input_ids) + self.encoder_positions.weight[:input_ids.shape[1]]
        x = self.encoder_norm(x)

        encoder_outputs = self.encoder(x, attention_mask)
        encoder_outputs = self.pre_decoder_layer(encoder_outputs)
        encoder_outputs = encoder_outputs.view(
            encoder_outputs.shape[:2] + (self.decoder_layers, 2, self.encoder_heads, -1)).transpose(1, 4).contiguous()

        x = self.embeddings(decoder_input_ids) + self.decoder_positions.weight[:input_ids.shape[1]]
        x = self.decoder_norm(x)

        return self.decoder(
            tensor=x, decoder_mask=decoder_padding_mask, encoder_state=encoder_outputs, encoder_mask=attention_mask
        )[0]


class BartForSequenceClassification(Module):
    def __init__(
            self, vocab_size: int, padding_idx: int, dim: int, max_position_embeddings: int,
            encoder_layers: int, encoder_heads: int, encoder_ffn_dim: int,
            decoder_layers: int, decoder_heads: int, decoder_ffn_dim: int, eos_token_id: int, num_labels: int
    ):
        super().__init__()
        self.transformer = BartModel(
            vocab_size=vocab_size, padding_idx=padding_idx, eos_token_id=eos_token_id, dim=dim,
            max_position_embeddings=max_position_embeddings, encoder_layers=encoder_layers,
            encoder_heads=encoder_heads, encoder_ffn_dim=encoder_ffn_dim, decoder_layers=decoder_layers,
            decoder_heads=decoder_heads, decoder_ffn_dim=decoder_ffn_dim
        )
        self.classification_head = TanhHead(dim, dim, num_labels)

        self.eos_token_id = eos_token_id

    def replace_encoder_keys(self, state_dict: dict):
        encoder_old_names = {
            'attention.out_lin': 'self_attn.out_proj',
            'norm1': 'self_attn_layer_norm',
            'norm2': 'final_layer_norm',
            'ffn.lin1': 'fc1',
            'ffn.lin2': 'fc2'
        }

        for layer_ind in range(len(self.transformer.encoder.layers)):
            prefix = f'model.encoder.layers.{layer_ind}.'
            for w in ('weight', 'bias'):
                state_dict[f'{prefix}attention.pre_attention.{w}'] = transposed_cat([
                    state_dict.pop(f'{prefix}self_attn.{k}_proj.{w}') for k in ('q', 'k', 'v')
                ])
                for new_name, old_name in encoder_old_names.items():
                    state_dict[f'{prefix}{new_name}.{w}'] = state_dict.pop(f'{prefix}{old_name}.{w}')

    def replace_decoder_keys(self, state_dict: dict):
        decoder_old_names = {
            'encoder_attention.q_lin': 'encoder_attn.q_proj',
            'encoder_attention.out_lin': 'encoder_attn.out_proj',
            'self_attention.out_lin': 'self_attn.out_proj',
            'ffn.lin1': 'fc1',
            'ffn.lin2': 'fc2',
            'norm1': 'self_attn_layer_norm',
            'norm2': 'encoder_attn_layer_norm',
            'norm3': 'final_layer_norm'
        }

        for layer_ind in range(len(self.transformer.encoder.layers)):
            prefix = f'model.decoder.layers.{layer_ind}.'
            for w in ('weight', 'bias'):
                state_dict[f'{prefix}self_attention.pre_attention.{w}'] = transposed_cat([
                    state_dict.pop(f'{prefix}self_attn.{k}_proj.{w}') for k in ('q', 'k', 'v')
                ])
                for new_name, old_name in decoder_old_names.items():
                    state_dict[f'{prefix}{new_name}.{w}'] = state_dict.pop(f'{prefix}{old_name}.{w}')

    def load_state_dict(self, state_dict, strict: bool = True):
        keys_to_delete = [
            'model.encoder.version', 'model.decoder.version', 'model.encoder.embed_tokens.weight',
            'model.decoder.embed_tokens.weight']
        for key in keys_to_delete:
            del state_dict[key]
        old_names = {
            'model.encoder_positions.weight': 'model.encoder.embed_positions.weight',
            'model.encoder_norm.weight': 'model.encoder.layernorm_embedding.weight',
            'model.encoder_norm.bias': 'model.encoder.layernorm_embedding.bias',
            'model.decoder_positions.weight': 'model.decoder.embed_positions.weight',
            'model.decoder_norm.weight': 'model.decoder.layernorm_embedding.weight',
            'model.decoder_norm.bias': 'model.decoder.layernorm_embedding.bias'
        }

        for new_name, old_name in old_names.items():
            state_dict[new_name] = state_dict.pop(old_name)

        state_dict['model.encoder_positions.weight'] = state_dict['model.encoder_positions.weight'][2:]
        state_dict['model.decoder_positions.weight'] = state_dict['model.decoder_positions.weight'][2:]

        for p in ('weight', 'bias'):
            state_dict[f'model.pre_decoder_layer.{p}'] = torch.cat(
                [state_dict.pop(f'model.decoder.layers.{ind}.encoder_attn.{name}_proj.{p}')
                 for ind in range(len(self.transformer.decoder.layers)) for name in 'kv'], dim=0)

        self.replace_encoder_keys(state_dict)
        self.replace_decoder_keys(state_dict)

        module_names = {'embeddings': 'shared', 'transformer': 'model'}
        for key, value in list(state_dict.items()):
            new_key = key
            for name, old_name in module_names.items():
                new_key = new_key.replace(f'{old_name}.', f'{name}.')
            state_dict[new_key] = state_dict.pop(key)

        super().load_state_dict(state_dict, strict)

    @classmethod
    def from_pretrained(cls, model_path: str) -> 'BartForSequenceClassification':
        model = cls(
            vocab_size=50265, padding_idx=1, dim=1024, max_position_embeddings=1024,
            encoder_layers=12, encoder_heads=16, encoder_ffn_dim=4096,
            decoder_layers=12, decoder_heads=16, decoder_ffn_dim=4096, num_labels=3, eos_token_id=2
        )
        model.load_state_dict(state_dict=torch.load(os.path.join(model_path, 'pytorch_model.bin'), map_location='cpu'))
        model.eval()
        for p in model.parameters():
            p.requires_grad = False
        return model.half()

    def forward(self, input_ids: Tensor, attention_mask: Tensor) -> Tensor:
        tensor = self.transformer(input_ids, attention_mask=attention_mask)
        eos_mask = input_ids.eq(self.eos_token_id)
        bsz, _, dim = tensor.shape
        tensor = tensor[eos_mask].view(bsz, -1, dim)[:, -1]
        return self.classification_head(tensor)



import torch
from torch.nn.functional import gelu

from .base_modules import Linear


class TransformerFFN(torch.nn.Module):
    def __init__(self, dim: int, dim_hidden: int):
        super(TransformerFFN, self).__init__()
        self.lin1 = Linear(dim, dim_hidden)
        self.lin2 = Linear(dim_hidden, dim)

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        tensor = self.lin1(tensor)
        tensor = gelu(tensor)
        return self.lin2(tensor)



import torch

from .time_distributed import TimeDistributed
from .coref_utils import batched_index_select, batched_span_select, masked_softmax, weighted_sum


class SpanExtractor(torch.nn.Module):
    def __init__(self, num_width_embeddings: int, span_width_embedding_dim: int):
        super().__init__()

        self._span_width_embedding = torch.nn.Embedding(
            num_embeddings=num_width_embeddings, embedding_dim=span_width_embedding_dim)

    def forward(self, sequence_tensor: torch.Tensor, span_indices: torch.Tensor) -> torch.Tensor:
        span_starts, span_ends = [index.squeeze(-1) for index in span_indices.split(1, dim=-1)]

        start_embeddings = batched_index_select(sequence_tensor, span_starts)
        end_embeddings = batched_index_select(sequence_tensor, span_ends)
        span_width_embeddings = self._span_width_embedding(span_ends - span_starts)
        return torch.cat([start_embeddings, end_embeddings, span_width_embeddings], -1)


class SelfAttentiveSpanExtractor(torch.nn.Module):
    def __init__(self, input_dim: int) -> None:
        super().__init__()
        self._global_attention = TimeDistributed(torch.nn.Linear(input_dim, 1))

    def forward(self, sequence_tensor: torch.Tensor, span_indices: torch.Tensor) -> torch.Tensor:
        # shape (batch_size, sequence_length, 1)
        global_attention_logits = self._global_attention(sequence_tensor)

        # shape (batch_size, sequence_length, embedding_dim + 1)
        concat_tensor = torch.cat([sequence_tensor, global_attention_logits], -1)

        concat_output, span_mask = batched_span_select(concat_tensor, span_indices)

        # Shape: (batch_size, num_spans, max_batch_span_width, embedding_dim)
        span_embeddings = concat_output[:, :, :, :-1]
        # Shape: (batch_size, num_spans, max_batch_span_width)
        span_attention_logits = concat_output[:, :, :, -1]

        # Shape: (batch_size, num_spans, max_batch_span_width)
        span_attention_weights = masked_softmax(span_attention_logits, span_mask)

        # Do a weighted sum of the embedded spans with
        # respect to the normalised attention distributions.
        # Shape: (batch_size, num_spans, embedding_dim)
        return weighted_sum(span_embeddings, span_attention_weights)



import logging
import math
from typing import Tuple

import torch
from torch.nn.functional import pad

from .coref_utils import batched_index_select, batched_span_select
from ..bert import BertModel

logger = logging.getLogger(__name__)


class TransformerEmbedder(torch.nn.Module):
    def __init__(self, max_length: int):
        super().__init__()

        self.model = BertModel(
            vocab_size=28996, dim=1024, padding_idx=0, max_position_embeddings=512, ffn_size=4096, n_heads=16,
            n_layers=24
        )

        self._max_length = max_length

        self.output_dim = 1024

        self._num_added_start_tokens = 1
        self._num_added_end_tokens = 1
        self._num_added_tokens = self._num_added_start_tokens + self._num_added_end_tokens

    def get_output_dim(self):
        return self.output_dim

    def forward(
            self, token_ids: torch.Tensor, offsets: torch.Tensor, segment_concat_mask: torch.Tensor) -> torch.Tensor:
        fold_long_sequences = token_ids.shape[1] > self._max_length
        batch_size, num_segment_concat_wordpieces = token_ids.shape
        if fold_long_sequences:
            token_ids, segment_concat_mask = self._fold_long_sequences(token_ids, segment_concat_mask)

        # Shape: [batch_size, num_wordpieces, embedding_size],
        # or if self._max_length is not None:
        # [batch_size * num_segments, self._max_length, embedding_size]
        embeddings = self.model(input_ids=token_ids, attention_mask=segment_concat_mask)

        if fold_long_sequences:
            embeddings = self._unfold_long_sequences(
                embeddings, segment_concat_mask, batch_size, num_segment_concat_wordpieces)

        span_embeddings, span_mask = batched_span_select(embeddings.contiguous(), offsets)

        span_mask = span_mask.unsqueeze(-1)

        # Shape: (batch_size, num_orig_tokens, max_span_length, embedding_size)
        span_embeddings *= span_mask  # zero out paddings

        # Sum over embeddings of all sub-tokens of a word
        # Shape: (batch_size, num_orig_tokens, embedding_size)
        span_embeddings_sum = span_embeddings.sum(2)

        # Shape (batch_size, num_orig_tokens)
        span_embeddings_len = span_mask.sum(2)

        # Find the average of sub-tokens embeddings by dividing `span_embedding_sum` by `span_embedding_len`
        # Shape: (batch_size, num_orig_tokens, embedding_size)
        orig_embeddings = span_embeddings_sum / torch.clamp_min(span_embeddings_len, 1)

        # All the places where the span length is zero, write in zeros.
        orig_embeddings[(span_embeddings_len == 0).expand(orig_embeddings.shape)] = 0

        return orig_embeddings

    def fold(self, tensor: torch.Tensor, length_to_pad: int) -> torch.Tensor:
        # Shape: [batch_size, num_segments * self._max_length]
        tensor = pad(tensor, [0, length_to_pad], value=0.)
        # Shape: [batch_size * num_segments, self._max_length]
        return tensor.reshape(-1, self._max_length)

    def _fold_long_sequences(self, token_ids: torch.Tensor, mask: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        We fold 1D sequences (for each element in batch), returned by `PretrainedTransformerIndexer`
        that are in reality multiple segments concatenated together, to 2D tensors, e.g.

        [ [CLS] A B C [SEP] [CLS] D E [SEP] ]
        -> [ [ [CLS] A B C [SEP] ], [ [CLS] D E [SEP] [PAD] ] ]
        The [PAD] positions can be found in the returned `mask`.

        # Parameters

        token_ids: `torch.LongTensor`
            Shape: `[batch_size, num_segment_concat_wordpieces]`.
            num_segment_concat_wordpieces is num_wordpieces plus special tokens inserted in the
            middle, i.e. the length of: "[CLS] A B C [SEP] [CLS] D E F [SEP]" (see indexer logic).
        mask: `torch.BoolTensor`
            Shape: `[batch_size, num_segment_concat_wordpieces]`.
            The mask for the concatenated segments of wordpieces. The same as `segment_concat_mask`
            in `forward()`.
        type_ids: `Optional[torch.LongTensor]`
            Shape: [batch_size, num_segment_concat_wordpieces].

        # Returns:

        token_ids: `torch.LongTensor`
            Shape: [batch_size * num_segments, self._max_length].
        mask: `torch.BoolTensor`
            Shape: [batch_size * num_segments, self._max_length].
        """
        num_segment_concat_wordpieces = token_ids.shape[1]
        num_segments = math.ceil(num_segment_concat_wordpieces / self._max_length)  # type: ignore
        padded_length = num_segments * self._max_length  # type: ignore
        length_to_pad = padded_length - num_segment_concat_wordpieces

        return self.fold(token_ids, length_to_pad), self.fold(mask, length_to_pad)

    def _unfold_long_sequences(
        self,
        embeddings: torch.Tensor,
        mask: torch.Tensor,
        batch_size: int,
        num_segment_concat_wordpieces: int,
    ) -> torch.Tensor:
        """
        We take 2D segments of a long sequence and flatten them out to get the whole sequence
        representation while remove unnecessary special tokens.

        [ [ [CLS]_emb A_emb B_emb C_emb [SEP]_emb ], [ [CLS]_emb D_emb E_emb [SEP]_emb [PAD]_emb ] ]
        -> [ [CLS]_emb A_emb B_emb C_emb D_emb E_emb [SEP]_emb ]

        We truncate the start and end tokens for all segments, recombine the segments,
        and manually add back the start and end tokens.

        # Parameters

        embeddings: `torch.FloatTensor`
            Shape: [batch_size * num_segments, self._max_length, embedding_size].
        mask: `torch.BoolTensor`
            Shape: [batch_size * num_segments, self._max_length].
            The mask for the concatenated segments of wordpieces. The same as `segment_concat_mask`
            in `forward()`.
        batch_size: `int`
        num_segment_concat_wordpieces: `int`
            The length of the original "[ [CLS] A B C [SEP] [CLS] D E F [SEP] ]", i.e.
            the original `token_ids.size(1)`.

        # Returns:

        embeddings: `torch.FloatTensor`
            Shape: [batch_size, self._num_wordpieces, embedding_size].
        """

        device = embeddings.device
        num_segments = int(embeddings.shape[0] / batch_size)
        embedding_size = embeddings.shape[2]

        # We want to remove all segment-level special tokens but maintain sequence-level ones
        num_wordpieces = num_segment_concat_wordpieces - (num_segments - 1) * self._num_added_tokens

        embeddings = embeddings.reshape(batch_size, num_segments * self._max_length, embedding_size)
        mask = mask.reshape(batch_size, num_segments * self._max_length)  # type: ignore
        # Shape: (batch_size,)
        seq_lengths = mask.sum(-1)
        # Shape: (batch_size, self._num_added_end_tokens); this is a broadcast op
        end_token_indices = seq_lengths.unsqueeze(-1) - torch.arange(self._num_added_end_tokens, device=device) - 1

        # Shape: (batch_size, self._num_added_start_tokens, embedding_size)
        start_token_embeddings = embeddings[:, : self._num_added_start_tokens]
        # Shape: (batch_size, self._num_added_end_tokens, embedding_size)
        end_token_embeddings = batched_index_select(embeddings, end_token_indices)

        embeddings = embeddings.reshape(batch_size, num_segments, self._max_length, embedding_size)
        embeddings = embeddings[:, :, self._num_added_start_tokens : embeddings.shape[2] - self._num_added_end_tokens]
        embeddings = embeddings.reshape(batch_size, -1, embedding_size)  # flatten

        # Now try to put end token embeddings back which is a little tricky.

        # The number of segment each sequence spans, excluding padding. Mimicking ceiling operation.
        # Shape: (batch_size,)
        num_effective_segments = (seq_lengths + self._max_length - 1) // self._max_length
        # The number of indices that end tokens should shift back.
        num_removed_non_end_tokens = num_effective_segments * self._num_added_tokens - self._num_added_end_tokens

        # Shape: (batch_size, self._num_added_end_tokens)
        end_token_indices -= num_removed_non_end_tokens.unsqueeze(-1)
        assert (end_token_indices >= self._num_added_start_tokens).all()
        # Add space for end embeddings
        embeddings = torch.cat([embeddings, torch.zeros_like(end_token_embeddings)], 1)
        # Add end token embeddings back
        embeddings.scatter_(1, end_token_indices.unsqueeze(-1).expand_as(end_token_embeddings), end_token_embeddings)

        # Now put back start tokens. We can do this before putting back end tokens, but then
        # we need to change `num_removed_non_end_tokens` a little.
        embeddings = torch.cat([start_token_embeddings, embeddings], 1)

        # Truncate to original length
        return embeddings[:, :num_wordpieces]



from ._version import __version__

from .models import (
    BartForSequenceClassification, Blender, DistilBertForMaskedLM,
    ElectraForSequenceClassification, RobertaForQuestionAnswering
)

from .multiprocess_api import ClientAPI, ModelDeploymentConfig

from .run_model_fleet import run_model_fleet



import os
import torch

from ..modules import TransformerEncoder, MaskEmbeddings


class BertModel(torch.nn.Module):
    def __init__(
            self, vocab_size: int, dim: int, padding_idx: int,
            max_position_embeddings: int, ffn_size: int, n_heads: int, n_layers: int):
        super().__init__()

        self.embeddings = MaskEmbeddings(
            vocab_size=vocab_size, dim=dim, padding_idx=padding_idx,
            max_position_embeddings=max_position_embeddings
        )
        self.encoder = TransformerEncoder(
            n_heads=n_heads, embedding_size=dim, ffn_size=ffn_size, n_layers=n_layers)

    def forward(self, input_ids: torch.Tensor, attention_mask: torch.Tensor) -> torch.Tensor:
        embedding_output = self.embeddings(input_ids=input_ids)
        return self.encoder.forward(embedding_output, attention_mask.to(torch.bool))

    @classmethod
    def from_pretrained(cls, model_path: str) -> 'BertModel':
        model = cls(
            vocab_size=28996, dim=1024, padding_idx=0, max_position_embeddings=512,
            ffn_size=4096, n_heads=16, n_layers=24
        )
        model.load_state_dict(state_dict=torch.load(os.path.join(model_path, 'pytorch_model.bin'), map_location='cpu'))
        model.eval()
        for p in model.parameters():
            p.requires_grad = False
        return model.half()



from typing import Optional, Tuple, List

import torch
from torch.nn import Module, ModuleList

from .base_modules import LayerNorm
from .attention import MultiHeadAttention, DecoderEncoderAttention
from .feed_forward import TransformerFFN


class TransformerDecoderLayer(Module):
    def __init__(self, n_heads: int, embedding_size: int, ffn_size: int, blender_norm: bool):
        super().__init__()
        self.dim = embedding_size

        self.self_attention = MultiHeadAttention(n_heads, embedding_size)
        self.norm1 = LayerNorm(embedding_size)

        self.encoder_attention = DecoderEncoderAttention(n_heads, embedding_size)
        self.norm2 = LayerNorm(embedding_size)

        self.ffn = TransformerFFN(embedding_size, ffn_size)
        self.norm3 = LayerNorm(embedding_size)

        self.blender_norm = blender_norm

    def forward(
            self, tensor, decoder_mask: torch.Tensor, encoder_state: torch.Tensor, encoder_mask: torch.Tensor,
            incr_state: Optional[torch.Tensor] = None, get_incr_state: bool = False
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        residual = tensor
        if self.blender_norm:
            tensor = self.norm1(tensor)
        tensor, incr_state = self.self_attention(
            query=tensor, mask=decoder_mask, incr_state=incr_state, get_incr_state=get_incr_state)
        tensor = tensor + residual
        if not self.blender_norm:
            tensor = self.norm1(tensor)
        residual = tensor
        if self.blender_norm:
            tensor = self.norm2(tensor)
        tensor = self.encoder_attention(
            query=tensor, key=encoder_state[:, 0], value=encoder_state[:, 1], mask=encoder_mask)
        tensor = tensor + residual
        if not self.blender_norm:
            tensor = self.norm2(tensor)
        residual = tensor
        if self.blender_norm:
            tensor = self.norm3(tensor)
        tensor = self.ffn(tensor)
        tensor = residual + tensor
        if not self.blender_norm:
            tensor = self.norm3(tensor)
        return tensor, incr_state


class TransformerDecoder(Module):
    def __init__(self, n_heads: int, n_layers: int, embedding_size: int, ffn_size: int, blender_norm: bool = False):
        super().__init__()

        self.layers = ModuleList(
            TransformerDecoderLayer(n_heads, embedding_size, ffn_size, blender_norm) for _ in range(n_layers))
        self.n_heads = n_heads

    def process_decoder_mask(self, decoder_mask: torch.Tensor, decoder_length: int) -> torch.Tensor:
        bsz, seq_len = decoder_mask.shape
        mask = torch.tril(torch.ones(seq_len, seq_len, dtype=torch.bool, device=decoder_mask.device))
        mask = mask.unsqueeze(0)[:, -decoder_length:] & decoder_mask.unsqueeze(1)
        return ~mask.repeat_interleave(repeats=self.n_heads, dim=0).expand(-1, decoder_length, -1)

    def forward(
            self, tensor: torch.Tensor, decoder_mask: torch.Tensor,
            encoder_mask: torch.Tensor, encoder_state: torch.Tensor,
            incr_state: Optional[torch.Tensor] = None, get_incr_state: bool = True
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:

        decoder_length = tensor.shape[1]
        decoder_mask = self.process_decoder_mask(decoder_mask, decoder_length=decoder_length)
        encoder_mask = ~encoder_mask.repeat_interleave(
            repeats=self.n_heads, dim=0).unsqueeze(1).expand(-1, decoder_length, -1)

        encoder_state = encoder_state.flatten(0, 1)

        new_incr_states: List[torch.Tensor] = []
        for ind, layer in enumerate(self.layers):
            tensor, new_layer_state = layer.forward(
                tensor=tensor,
                decoder_mask=decoder_mask,
                encoder_state=encoder_state[:, ind],
                encoder_mask=encoder_mask,
                incr_state=incr_state[:, :, :, :, ind] if incr_state is not None else None,
                get_incr_state=get_incr_state
            )
            if get_incr_state:
                assert new_layer_state is not None
                new_incr_states.append(new_layer_state)

        return tensor, torch.stack(new_incr_states, dim=-1) if get_incr_state else None



from typing import Optional, Union, Tuple

import torch


class Linear(torch.nn.Linear):
    def __init__(self, in_features: int, out_features: int, bias: bool = True):
        torch.nn.Module.__init__(self)
        self.in_features = in_features
        self.out_features = out_features
        self.weight = torch.nn.Parameter(torch.empty(out_features, in_features, dtype=torch.float))
        if bias:
            self.bias = torch.nn.Parameter(torch.empty(out_features, dtype=torch.float))
        else:
            self.register_parameter('bias', None)


class Embedding(torch.nn.Embedding):
    def __init__(self, num_embeddings: int, embedding_dim: int, padding_idx: Optional[int] = None):
        torch.nn.Module.__init__(self)
        self.weight = torch.nn.Parameter(torch.empty(num_embeddings, embedding_dim, dtype=torch.float))
        self.num_embeddings = num_embeddings
        self.embedding_dim = embedding_dim
        if padding_idx is not None:
            if padding_idx > 0:
                assert padding_idx < self.num_embeddings, 'Padding_idx must be within num_embeddings'
            elif padding_idx < 0:
                assert padding_idx >= -self.num_embeddings, 'Padding_idx must be within num_embeddings'
                padding_idx += self.num_embeddings
        self.padding_idx = padding_idx
        self.max_norm = None
        self.norm_type = 2.
        self.scale_grad_by_freq = False
        self.sparse = False


class LayerNorm(torch.nn.LayerNorm):
    def __init__(
            self, normalized_shape: Union[int, Tuple[int, ...]], eps: float = 1e-5, elementwise_affine: bool = True
    ):
        torch.nn.Module.__init__(self)
        self.normalized_shape = (normalized_shape,) if isinstance(normalized_shape, int) else normalized_shape
        self.eps = eps
        self.elementwise_affine = elementwise_affine
        if elementwise_affine:
            self.weight = torch.nn.Parameter(torch.empty(normalized_shape, dtype=torch.float))
            self.bias = torch.nn.Parameter(torch.empty(normalized_shape, dtype=torch.float))
        else:
            self.register_parameter('weight', None)
            self.register_parameter('bias', None)



from typing import Any, Type, Dict

import multiprocessing
import torch

from .multiprocess_api import ModelServer, ModelDeploymentConfig


def run_server(
        model_class: Type[torch.nn.Module], path: str, model_kwargs: Dict[str, Any], device: str, port: int,
        torchscript: bool, verbose: bool
):
    model = model_class.from_pretrained(path, **model_kwargs).to(device)
    if torchscript:
        model = torch.jit.script(model)
    server = ModelServer(port=port, model=model, verbose=verbose)
    try:
        server.run()
    finally:
        server.close()


def run_model_fleet(model_deployments: Dict[str, ModelDeploymentConfig], verbose: bool = False):
    for model_config in model_deployments.values():
        def run_worker_server(worker, verbose):
            run_server(
                model_class=model_config.model_class, path=model_config.path, model_kwargs=model_config.model_kwargs,
                device=f'cuda:{model_config.allocations[worker]}', port=model_config.port + worker,
                torchscript=model_config.use_torchscript, verbose=verbose
            )
        processes = [
            multiprocessing.Process(target=run_worker_server, args=(worker, verbose))
            for worker in range(len(model_config.allocations))
        ]
        for process in processes:
            process.start()



from typing import Optional

import torch

from .base_modules import Embedding, LayerNorm


class Embeddings(torch.nn.Module):
    def __init__(self, vocab_size: int, dim: int, padding_idx: int, max_position_embeddings: int):
        super().__init__()
        self.word_embeddings = Embedding(vocab_size, dim, padding_idx=padding_idx)
        self.position_embeddings = Embedding(max_position_embeddings, dim)
        self.LayerNorm = LayerNorm(dim)

    def forward(self, input_ids: torch.Tensor) -> torch.Tensor:
        embeddings = self.word_embeddings(input_ids) + self.position_embeddings.weight[:input_ids.shape[1]]
        return self.LayerNorm(embeddings)


class MaskEmbeddings(Embeddings):
    def __init__(self, vocab_size: int, dim: int, padding_idx: int, max_position_embeddings: int):
        super().__init__(vocab_size, dim, padding_idx, max_position_embeddings)
        self.padding_idx = padding_idx

    @staticmethod
    def create_position_ids_from_input_ids(input_ids: torch.Tensor, padding_idx: int) -> torch.Tensor:
        mask = input_ids.ne(padding_idx)
        return torch.cumsum(mask, dim=1) * mask + padding_idx

    def forward(self, input_ids: torch.Tensor) -> torch.Tensor:
        position_ids = self.create_position_ids_from_input_ids(input_ids, self.padding_idx)
        embedded = self.word_embeddings(input_ids) + self.position_embeddings(position_ids)
        return self.LayerNorm(embedded)


class EmbeddingsWithTokenTypes(torch.nn.Module):
    def __init__(
            self, vocab_size: int, token_types: int, embedding_size: int, pad_token_id: int,
            max_position_embeddings: int
    ):
        super().__init__()
        self.word_embeddings = Embedding(vocab_size, embedding_size, padding_idx=pad_token_id)
        self.position_embeddings = Embedding(max_position_embeddings, embedding_size)
        self.token_type_embeddings = Embedding(token_types, embedding_size)
        self.LayerNorm = LayerNorm(embedding_size)

    def forward(self, input_ids: torch.Tensor, token_type_ids: Optional[torch.Tensor]) -> torch.Tensor:
        embeddings = self.word_embeddings(input_ids) + self.position_embeddings.weight[:input_ids.shape[1]] + (
            self.token_type_embeddings.weight[0] if token_type_ids is None
            else self.token_type_embeddings(token_type_ids))
        return self.LayerNorm(embeddings)



from typing import Any, Optional, Union, Tuple, List, Dict, Iterable

import select
import socket
import io
import time
from operator import itemgetter

import torch.jit

from .configs import Config
from ..utils import logger

from .socket_connection import SocketConnection


class ServerSocket(SocketConnection):
    def read_buffer(self) -> Optional[bytes]:
        response = bytes()
        while True:
            buffer = self.connection.recv(self.message_size)
            if not buffer:
                raise ConnectionError('No response, other side probably died')
            response += buffer
            if response.endswith(Config.request_cancelled):
                return None
            if response.endswith(Config.end_delimiter):
                return response.split(Config.end_delimiter)[-2]

    def receive_inputs(self, device: str) -> Optional[Tuple[bytes, Any]]:
        try:
            request_message = self.connection.recv(self.message_size)
            if not request_message:
                raise ConnectionError('No response, other side probably died')
        except ConnectionError:
            self.alive = False
            raise
        request_message = request_message.split(Config.request_cancelled)[-1]
        if not request_message:
            return None
        self.connection.send(Config.request_signal)
        response = self.read_buffer()
        if response is None:
            return None
        byte_inputs = io.BytesIO(response[Config.message_id_len:])
        return response[:Config.message_id_len], torch.load(byte_inputs, map_location=device)


class SessionState:
    def __init__(self, connection: SocketConnection, message_id: bytes, inputs: dict):
        self.connection = connection
        self.message_id = message_id
        self.inputs = inputs

    def send_output(self, output):
        try:
            self.connection.send_message(self.message_id, output)
        except ConnectionError:
            pass

    def send_error(self, e: Exception):
        try:
            self.connection.send_error(self.message_id, str(e))
        except ConnectionError:
            pass


class ModelServer:
    __slots__ = ('model', 'device', 'server_socket', 'verbose')

    def __init__(self, port: int, model: torch.nn.Module, verbose: bool):
        logger.info(f'{type(model).__name__} model server starting on port {port}')
        self.model = model
        self.device = next(self.model.parameters()).device
        self.server_socket = self.create_server_socket(port)
        self.verbose = verbose
        logger.info(f'{type(model).__name__} model server started on port {port}')

    @staticmethod
    def create_server_socket(port: int) -> socket.socket:
        server_socket = socket.socket()
        while True:
            try:
                server_socket.bind((Config.host, port))
                break
            except OSError:
                time.sleep(5)
        server_socket.listen(10)
        return server_socket

    def close(self):
        self.server_socket.close()

    @staticmethod
    def pad_and_concat(tensors: List[torch.Tensor], pad: int = 0) -> torch.Tensor:
        sample_tensor = tensors[0]
        if len(tensors) == 1:
            return sample_tensor

        max_len = max([tensor.shape[1] for tensor in tensors])

        total_batch_size = 0
        for tensor in tensors:
            total_batch_size += tensor.shape[0]
        output_tensor = torch.empty((total_batch_size, max_len) + sample_tensor.shape[2:],
                                    dtype=sample_tensor.dtype, device=sample_tensor.device).fill_(pad)
        cur_ind = 0
        for tensor in tensors:
            new_ind = cur_ind + tensor.shape[0]
            output_tensor[cur_ind: new_ind, :tensor.shape[1]] = tensor
            cur_ind = new_ind

        return output_tensor

    @staticmethod
    def split_tensor_output(outputs: torch.Tensor, sizes: List[int]) -> Iterable[torch.Tensor]:
        outputs = outputs.detach().to('cpu')
        last_ind = 0
        for size in sizes:
            new_ind = last_ind + size
            yield outputs[last_ind: new_ind]
            last_ind = new_ind

    @staticmethod
    def split_dict_output(
            outputs: Dict[str, Union[list, torch.Tensor]], sizes: List[int]
    ) -> Iterable[Dict[str, Union[list, torch.Tensor]]]:
        outputs = {
            key: value.detach().to('cpu') if isinstance(value, torch.Tensor) else value
            for key, value in outputs.items()
        }
        last_ind = 0
        for size in sizes:
            new_ind = last_ind + size
            yield {key: value[last_ind: new_ind] for key, value in outputs.items()}
            last_ind = new_ind

    def split_output(self, outputs, sizes: List[int]) -> Iterable:
        return (self.split_tensor_output(outputs, sizes) if isinstance(outputs, torch.Tensor)
                else self.split_dict_output(outputs, sizes))

    def run(self):
        connections: List[Union[SocketConnection, socket.socket]] = []

        while True:
            connections = [connection for connection in connections if connection.alive]

            new_connections = select.select(connections + [self.server_socket], [], [], None)[0]

            if self.server_socket in new_connections:
                connection, address = self.server_socket.accept()
                connection = ServerSocket(connection, Config.server_message_size)
                logger.debug(f'Added connection, fileno: {connection.fileno()}')
                connections.append(connection)
                continue

            session_states: List[SessionState] = []
            for connection in new_connections:
                try:
                    inputs = connection.receive_inputs(self.device)
                except ConnectionError as e:
                    logger.warning(f'ConnectionError:\n{e}')
                    continue
                if not inputs:
                    continue
                message_id, input_dict = inputs
                session_states.append(SessionState(connection, message_id, input_dict))

            if not session_states:
                continue

            inputs = [session_state.inputs for session_state in session_states]
            pad_map: Dict[str, int] = getattr(self.model, 'pad_map', {})

            try:
                keys = list(inputs[0].keys())
                sizes = [sample_inputs[keys[0]].shape[0] for sample_inputs in inputs]
                inputs = {
                    key: self.pad_and_concat(list(map(itemgetter(key), inputs)), pad_map.get(key, 0))
                    for key in keys
                }
                start_time = time.time() if self.verbose else None
                with torch.no_grad():
                    outputs = self.model.forward(**inputs)
                if self.verbose:
                    logger.debug(
                        f'Processed inputs of sizes: {sizes}\n'
                        f'Total size: {sum(sizes)}'
                        f'\nTime spent: {time.time() - start_time}')
                for session_state, output in zip(session_states, self.split_output(outputs, sizes)):
                    session_state.send_output(output)
            except Exception as e:
                logger.exception(e)
                for session_state in session_states:
                    session_state.send_error(e)



from typing import Tuple

import os
import math

import torch
from torch.nn.functional import relu

from allennlp.data import Vocabulary
from allennlp.models.model import Model
from allennlp.modules import FeedForward, GatedSum
from allennlp.modules import Seq2SeqEncoder, TimeDistributed, TextFieldEmbedder
from allennlp.modules.span_extractors import SelfAttentiveSpanExtractor, EndpointSpanExtractor
from allennlp.common import Params

from .coref_utils import (
    masked_topk, flatten_and_batch_shift_indices, batched_index_select, masked_softmax,
    weighted_sum, bucket_values)
from ...utils import neg_inf


class CoreferencePredictor(Model):
    def __init__(
            self,
            vocab: Vocabulary,
            text_field_embedder: TextFieldEmbedder,
            context_layer: Seq2SeqEncoder,
            mention_feedforward: FeedForward,
            antecedent_feedforward: FeedForward,
            feature_size: int,
            max_span_width: int,
            spans_per_word: float,
            max_antecedents: int,
            coarse_to_fine: bool = False,
            inference_order: int = 1,
            lexical_dropout: float = 0.2,
            **kwargs
    ) -> None:
        super().__init__(vocab, **kwargs)

        self._text_field_embedder = text_field_embedder
        self._context_layer = context_layer
        self._mention_feedforward = TimeDistributed(mention_feedforward)
        self._mention_scorer = TimeDistributed(
            torch.nn.Linear(mention_feedforward.get_output_dim(), 1)
        )
        self._antecedent_feedforward = TimeDistributed(antecedent_feedforward)
        self._antecedent_scorer = TimeDistributed(
            torch.nn.Linear(antecedent_feedforward.get_output_dim(), 1)
        )

        self._endpoint_span_extractor = EndpointSpanExtractor(
            context_layer.get_output_dim(),
            combination="x,y",
            num_width_embeddings=max_span_width,
            span_width_embedding_dim=feature_size,
            bucket_widths=False,
        )
        self._attentive_span_extractor = SelfAttentiveSpanExtractor(input_dim=text_field_embedder.get_output_dim())

        # 10 possible distance buckets.
        self._num_distance_buckets = 10
        self._distance_embedding = torch.nn.Embedding(
            embedding_dim=feature_size, num_embeddings=self._num_distance_buckets)

        self._spans_per_word = spans_per_word
        self._max_antecedents = max_antecedents

        self._coarse2fine_scorer = torch.nn.Linear(
            mention_feedforward.get_input_dim(), mention_feedforward.get_input_dim())
        self._inference_order = inference_order
        self._span_updating_gated_sum = GatedSum(mention_feedforward.get_input_dim())

        self.pad_map = {'spans': -1}

    @classmethod
    def from_pretrained(cls, model_path: str):
        model_params = Params.from_file(os.path.join(model_path, 'config.json'))['model']
        for key in ('type', 'initializer'):
            del model_params[key]
        model = cls.from_params(vocab=None, params=model_params, serialization_dir=model_path)
        model.load_state_dict(torch.load(os.path.join(model_path, 'weights.th'), map_location='cpu'), strict=True)
        model.eval()
        for p in model.parameters():
            p.requires_grad = False
        return model

    def forward(self, spans, token_ids, mask, offsets, segment_concat_mask):
        text = {'tokens': {
            'token_ids': token_ids, 'mask': mask, 'type_ids': None, 'offsets': offsets,
            'wordpiece_mask': None, 'segment_concat_mask': segment_concat_mask
        }}

        # Shape: (batch_size, document_length, embedding_size)
        text_embeddings = self._text_field_embedder(text)

        batch_size, num_spans = spans.shape[:2]
        document_length = text_embeddings.shape[1]

        # Shape: (batch_size, num_spans)
        span_mask = (spans[:, :, 0] >= 0).squeeze(-1)
        # SpanFields return -1 when they are used as padding. As we do
        # some comparisons based on span widths when we attend over the
        # span representations that we generate from these indices, we
        # need them to be <= 0. This is only relevant in edge cases where
        # the number of spans we consider after the pruning stage is >= the
        # total number of spans, because in this case, it is possible we might
        # consider a masked span.
        # Shape: (batch_size, num_spans, 2)
        spans = relu(spans.float()).long()

        # Shape: (batch_size, document_length, encoding_dim)
        contextualized_embeddings = text_embeddings * mask.unsqueeze(dim=-1)
        # Shape: (batch_size, num_spans, 2 * encoding_dim + feature_size)
        endpoint_span_embeddings = self._endpoint_span_extractor(contextualized_embeddings, spans)
        # Shape: (batch_size, num_spans, emebedding_size)
        attended_span_embeddings = self._attentive_span_extractor(text_embeddings, spans)

        # Shape: (batch_size, num_spans, emebedding_size + 2 * encoding_dim + feature_size)
        span_embeddings = torch.cat([endpoint_span_embeddings, attended_span_embeddings], -1)

        # Prune based on mention scores.
        num_spans_to_keep = min(int(math.floor(self._spans_per_word * document_length)), num_spans)

        # Shape: (batch_size, num_spans)
        span_mention_scores = self._mention_scorer(self._mention_feedforward(span_embeddings)).squeeze(-1)
        # Shape: (batch_size, num_spans) for all 3 tensors
        top_span_mention_scores, top_span_mask, top_span_indices = masked_topk(
            span_mention_scores, span_mask, num_spans_to_keep)

        # Shape: (batch_size * num_spans_to_keep)
        # torch.index_select only accepts 1D indices, but here
        # we need to select spans for each element in the batch.
        # This reformats the indices to take into account their
        # index into the batch. We precompute this here to make
        # the multiple calls to util.batched_index_select below more efficient.
        flat_top_span_indices = flatten_and_batch_shift_indices(top_span_indices, num_spans)

        # Compute final predictions for which spans to consider as mentions.
        # Shape: (batch_size, num_spans_to_keep, 2)
        top_spans = batched_index_select(spans, top_span_indices, flat_top_span_indices)
        # Shape: (batch_size, num_spans_to_keep, embedding_size)
        top_span_embeddings = batched_index_select(span_embeddings, top_span_indices, flat_top_span_indices)

        # Compute indices for antecedent spans to consider.
        max_antecedents = min(self._max_antecedents, num_spans_to_keep)

        # Now that we have our variables in terms of num_spans_to_keep, we need to
        # compare span pairs to decide each span's antecedent. Each span can only
        # have prior spans as antecedents, and we only consider up to max_antecedents
        # prior spans. So the first thing we do is construct a matrix mapping a span's
        # index to the indices of its allowed antecedents.

        # Once we have this matrix, we reformat our variables again to get embeddings
        # for all valid antecedents for each span. This gives us variables with shapes
        # like (batch_size, num_spans_to_keep, max_antecedents, embedding_size), which
        # we can use to make coreference decisions between valid span pairs.

        # Shape: (batch_size, num_spans_to_keep, max_antecedents) for all 4 tensors
        top_partial_coreference_scores, top_antecedent_mask, top_antecedent_offsets, top_antecedent_indices = self.\
            _coarse_to_fine_pruning(top_span_embeddings, top_span_mention_scores, top_span_mask, max_antecedents)

        flat_top_antecedent_indices = flatten_and_batch_shift_indices(top_antecedent_indices, num_spans_to_keep)

        # Shape: (batch_size, num_spans_to_keep, max_antecedents, embedding_size)
        top_antecedent_embeddings = batched_index_select(
            top_span_embeddings, top_antecedent_indices, flat_top_antecedent_indices)
        # Shape: (batch_size, num_spans_to_keep, 1 + max_antecedents)
        coreference_scores = self._compute_coreference_scores(
            top_span_embeddings,
            top_antecedent_embeddings,
            top_partial_coreference_scores,
            top_antecedent_mask,
            top_antecedent_offsets,
        )

        dummy_mask = torch.ones(
            (batch_size, num_spans_to_keep, 1), dtype=torch.bool, device=top_antecedent_mask.device)
        # Shape: (batch_size, num_spans_to_keep, 1 + max_antecedents,)
        top_antecedent_with_dummy_mask = torch.cat([dummy_mask, top_antecedent_mask], -1)
        # Shape: (batch_size, num_spans_to_keep, 1 + max_antecedents)
        attention_weight = masked_softmax(
            coreference_scores, top_antecedent_with_dummy_mask
        )
        # Shape: (batch_size, num_spans_to_keep, 1 + max_antecedents, embedding_size)
        top_antecedent_with_dummy_embeddings = torch.cat(
            [top_span_embeddings.unsqueeze(2), top_antecedent_embeddings], 2
        )
        # Shape: (batch_size, num_spans_to_keep, embedding_size)
        attended_embeddings = weighted_sum(top_antecedent_with_dummy_embeddings, attention_weight)
        # Shape: (batch_size, num_spans_to_keep, embedding_size)
        top_span_embeddings = self._span_updating_gated_sum(top_span_embeddings, attended_embeddings)

        # Shape: (batch_size, num_spans_to_keep, max_antecedents, embedding_size)
        top_antecedent_embeddings = batched_index_select(
            top_span_embeddings, top_antecedent_indices, flat_top_antecedent_indices
        )
        # Shape: (batch_size, num_spans_to_keep, 1 + max_antecedents)
        coreference_scores = self._compute_coreference_scores(
            top_span_embeddings,
            top_antecedent_embeddings,
            top_partial_coreference_scores,
            top_antecedent_mask,
            top_antecedent_offsets
        )

        # We now have, for each span which survived the pruning stage,
        # a predicted antecedent. This implies a clustering if we group
        # mentions which refer to each other in a chain.
        # Shape: (batch_size, num_spans_to_keep)
        _, predicted_antecedents = coreference_scores.max(2)
        # Subtract one here because index 0 is the "no antecedent" class,
        # so this makes the indices line up with actual spans if the prediction
        # is greater than -1.
        predicted_antecedents -= 1

        return {
            'top_spans': top_spans.detach().cpu(),
            'antecedent_indices': top_antecedent_indices.detach().cpu(),
            'predicted_antecedents': predicted_antecedents.detach().cpu()
        }

    def _coarse_to_fine_pruning(
        self,
        top_span_embeddings: torch.Tensor,
        top_span_mention_scores: torch.Tensor,
        top_span_mask: torch.Tensor,
        max_antecedents: int,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Generates antecedents for each span and prunes down to `max_antecedents`. This method
        prunes antecedents using a fast bilinar interaction score between a span and a candidate
        antecedent, and the highest-scoring antecedents are kept.

        # Parameters

        top_span_embeddings: `torch.FloatTensor`, required.
            The embeddings of the top spans.
            (batch_size, num_spans_to_keep, embedding_size).
        top_span_mention_scores: `torch.FloatTensor`, required.
            The mention scores of the top spans.
            (batch_size, num_spans_to_keep).
        top_span_mask: `torch.BoolTensor`, required.
            The mask for the top spans.
            (batch_size, num_spans_to_keep).
        max_antecedents: `int`, required.
            The maximum number of antecedents to keep for each span.

        # Returns

        top_partial_coreference_scores: `torch.FloatTensor`
            The partial antecedent scores for each span-antecedent pair. Computed by summing
            the span mentions scores of the span and the antecedent as well as a bilinear
            interaction term. This score is partial because compared to the full coreference scores,
            it lacks the interaction term
            `w * FFNN([g_i, g_j, g_i * g_j, features])`.
            `(batch_size, num_spans_to_keep, max_antecedents)`
        top_antecedent_mask: `torch.BoolTensor`
            The mask representing whether each antecedent span is valid. Required since
            different spans have different numbers of valid antecedents. For example, the first
            span in the document should have no valid antecedents.
            `(batch_size, num_spans_to_keep, max_antecedents)`
        top_antecedent_offsets: `torch.LongTensor`
            The distance between the span and each of its antecedents in terms of the number
            of considered spans (i.e not the word distance between the spans).
            `(batch_size, num_spans_to_keep, max_antecedents)`
        top_antecedent_indices: `torch.LongTensor`
            The indices of every antecedent to consider with respect to the top k spans.
            `(batch_size, num_spans_to_keep, max_antecedents)`
        """
        batch_size, num_spans_to_keep = top_span_embeddings.size()[:2]

        # Shape: (1, num_spans_to_keep, num_spans_to_keep)
        _, _, valid_antecedent_mask = self._generate_valid_antecedents(
            num_spans_to_keep, num_spans_to_keep, top_span_embeddings.device
        )

        mention_one_score = top_span_mention_scores.unsqueeze(1)
        mention_two_score = top_span_mention_scores.unsqueeze(2)
        bilinear_weights = self._coarse2fine_scorer(top_span_embeddings).transpose(1, 2)
        bilinear_score = torch.matmul(top_span_embeddings, bilinear_weights)
        # Shape: (batch_size, num_spans_to_keep, num_spans_to_keep); broadcast op
        partial_antecedent_scores = mention_one_score + mention_two_score + bilinear_score

        # Shape: (batch_size, num_spans_to_keep, num_spans_to_keep); broadcast op
        span_pair_mask = top_span_mask.unsqueeze(-1) & valid_antecedent_mask

        # Shape: (batch_size, num_spans_to_keep, max_antecedents) * 3
        top_partial_coreference_scores, top_antecedent_mask, top_antecedent_indices = masked_topk(
            partial_antecedent_scores, span_pair_mask, max_antecedents)

        top_span_range = torch.arange(num_spans_to_keep, dtype=torch.long, device=top_antecedent_mask.device)
        # Shape: (num_spans_to_keep, num_spans_to_keep); broadcast op
        valid_antecedent_offsets = top_span_range.unsqueeze(-1) - top_span_range.unsqueeze(0)

        top_antecedent_offsets = batched_index_select(
            valid_antecedent_offsets.unsqueeze(0)
            .expand(batch_size, num_spans_to_keep, num_spans_to_keep)
            .reshape(batch_size * num_spans_to_keep, num_spans_to_keep, 1),
            top_antecedent_indices.view(-1, max_antecedents),
        ).reshape(batch_size, num_spans_to_keep, max_antecedents)

        return top_partial_coreference_scores, top_antecedent_mask, top_antecedent_offsets, top_antecedent_indices

    def _compute_span_pair_embeddings(
            self, top_span_embeddings: torch.Tensor, antecedent_embeddings: torch.Tensor,
            antecedent_offsets: torch.Tensor) -> torch.Tensor:
        """
        Computes an embedding representation of pairs of spans for the pairwise scoring function
        to consider. This includes both the original span representations, the element-wise
        similarity of the span representations, and an embedding representation of the distance
        between the two spans.

        # Parameters

        top_span_embeddings : `torch.FloatTensor`, required.
            Embedding representations of the top spans. Has shape
            (batch_size, num_spans_to_keep, embedding_size).
        antecedent_embeddings : `torch.FloatTensor`, required.
            Embedding representations of the antecedent spans we are considering
            for each top span. Has shape
            (batch_size, num_spans_to_keep, max_antecedents, embedding_size).
        antecedent_offsets : `torch.IntTensor`, required.
            The offsets between each top span and its antecedent spans in terms
            of spans we are considering. Has shape (batch_size, num_spans_to_keep, max_antecedents).

        # Returns

        span_pair_embeddings : `torch.FloatTensor`
            Embedding representation of the pair of spans to consider. Has shape
            (batch_size, num_spans_to_keep, max_antecedents, embedding_size)
        """
        # Shape: (batch_size, num_spans_to_keep, max_antecedents, embedding_size)
        target_embeddings = top_span_embeddings.unsqueeze(2).expand_as(antecedent_embeddings)

        # Shape: (batch_size, num_spans_to_keep, max_antecedents, embedding_size)
        antecedent_distance_embeddings = self._distance_embedding(
            bucket_values(antecedent_offsets, num_total_buckets=self._num_distance_buckets))

        # Shape: (batch_size, num_spans_to_keep, max_antecedents, embedding_size)
        return torch.cat(
            [
                target_embeddings,
                antecedent_embeddings,
                antecedent_embeddings * target_embeddings,
                antecedent_distance_embeddings,
            ], -1)

    def _compute_coreference_scores(
        self,
        top_span_embeddings: torch.Tensor,
        top_antecedent_embeddings: torch.Tensor,
        top_partial_coreference_scores: torch.Tensor,
        top_antecedent_mask: torch.Tensor,
        top_antecedent_offsets: torch.Tensor,
    ) -> torch.Tensor:
        """
        Computes scores for every pair of spans. Additionally, a dummy label is included,
        representing the decision that the span is not coreferent with anything. For the dummy
        label, the score is always zero. For the true antecedent spans, the score consists of
        the pairwise antecedent score and the unary mention scores for the span and its
        antecedent. The factoring allows the model to blame many of the absent links on bad
        spans, enabling the pruning strategy used in the forward pass.

        # Parameters

        top_span_embeddings : `torch.FloatTensor`, required.
            Embedding representations of the kept spans. Has shape
            (batch_size, num_spans_to_keep, embedding_size)
        top_antecedent_embeddings: `torch.FloatTensor`, required.
            The embeddings of antecedents for each span candidate. Has shape
            (batch_size, num_spans_to_keep, max_antecedents, embedding_size)
        top_partial_coreference_scores : `torch.FloatTensor`, required.
            Sum of span mention score and antecedent mention score. The coarse to fine settings
            has an additional term which is the coarse bilinear score.
            (batch_size, num_spans_to_keep, max_antecedents).
        top_antecedent_mask : `torch.BoolTensor`, required.
            The mask for valid antecedents.
            (batch_size, num_spans_to_keep, max_antecedents).
        top_antecedent_offsets : `torch.FloatTensor`, required.
            The distance between the span and each of its antecedents in terms of the number
            of considered spans (i.e not the word distance between the spans).
            (batch_size, num_spans_to_keep, max_antecedents).

        # Returns

        coreference_scores : `torch.FloatTensor`
            A tensor of shape (batch_size, num_spans_to_keep, max_antecedents + 1),
            representing the unormalised score for each (span, antecedent) pair
            we considered.

        """
        # Shape: (batch_size, num_spans_to_keep, max_antecedents, embedding_size)
        span_pair_embeddings = self._compute_span_pair_embeddings(
            top_span_embeddings, top_antecedent_embeddings, top_antecedent_offsets
        )

        # Shape: (batch_size, num_spans_to_keep, max_antecedents)
        antecedent_scores = self._antecedent_scorer(self._antecedent_feedforward(span_pair_embeddings)).squeeze(-1)
        antecedent_scores += top_partial_coreference_scores
        antecedent_scores = antecedent_scores.masked_fill(
            ~top_antecedent_mask, neg_inf(antecedent_scores.dtype == torch.half))

        # Shape: (batch_size, num_spans_to_keep, 1)
        dummy_scores = antecedent_scores.new_zeros(antecedent_scores.shape[:2] + (1,))

        # Shape: (batch_size, num_spans_to_keep, max_antecedents + 1)
        return torch.cat([dummy_scores, antecedent_scores], -1)

    @staticmethod
    def _generate_valid_antecedents(
            num_spans_to_keep: int, max_antecedents: int, device: torch.device
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        This method generates possible antecedents per span which survived the pruning
        stage. This procedure is `generic across the batch`. The reason this is the case is
        that each span in a batch can be coreferent with any previous span, but here we
        are computing the possible `indices` of these spans. So, regardless of the batch,
        the 1st span _cannot_ have any antecedents, because there are none to select from.
        Similarly, each element can only predict previous spans, so this returns a matrix
        of shape (num_spans_to_keep, max_antecedents), where the (i,j)-th index is equal to
        (i - 1) - j if j <= i, or zero otherwise.

        # Parameters

        num_spans_to_keep : `int`, required.
            The number of spans that were kept while pruning.
        max_antecedents : `int`, required.
            The maximum number of antecedent spans to consider for every span.
        device : `int`, required.
            The CUDA device to use.

        # Returns

        valid_antecedent_indices : `torch.LongTensor`
            The indices of every antecedent to consider with respect to the top k spans.
            Has shape `(num_spans_to_keep, max_antecedents)`.
        valid_antecedent_offsets : `torch.LongTensor`
            The distance between the span and each of its antecedents in terms of the number
            of considered spans (i.e not the word distance between the spans).
            Has shape `(1, max_antecedents)`.
        valid_antecedent_mask : `torch.BoolTensor`
            The mask representing whether each antecedent span is valid. Required since
            different spans have different numbers of valid antecedents. For example, the first
            span in the document should have no valid antecedents.
            Has shape `(1, num_spans_to_keep, max_antecedents)`.
        """
        # Shape: (num_spans_to_keep, 1)
        target_indices = torch.arange(num_spans_to_keep, dtype=torch.long, device=device).unsqueeze(1)

        # Shape: (1, max_antecedents)
        valid_antecedent_offsets = (torch.arange(max_antecedents, dtype=torch.long, device=device) + 1).unsqueeze(0)

        # This is a broadcasted subtraction.
        # Shape: (num_spans_to_keep, max_antecedents)
        raw_antecedent_indices = target_indices - valid_antecedent_offsets

        # In our matrix of indices, the upper triangular part will be negative
        # because the offsets will be > the target indices. We want to mask these,
        # because these are exactly the indices which we don't want to predict, per span.
        # Shape: (1, num_spans_to_keep, max_antecedents)
        valid_antecedent_mask = (raw_antecedent_indices >= 0).unsqueeze(0)

        # Shape: (num_spans_to_keep, max_antecedents)
        valid_antecedent_indices = raw_antecedent_indices.clamp(min=0)
        return valid_antecedent_indices, valid_antecedent_offsets, valid_antecedent_mask



from typing import Optional

import os
import logging

import torch
from torch import Tensor
from torch.nn import Module
from torch.nn.functional import gelu

from ..modules import Linear, TransformerEncoder, EmbeddingsWithTokenTypes
from ..utils import transposed_cat


logger = logging.getLogger(__name__)


class ElectraModel(Module):
    def __init__(
            self, vocab_size: int, token_types: int, dim: int, pad_token_id: int,
            max_position_embeddings: int, ffn_size: int, n_heads: int, n_layers: int
    ):
        super().__init__()
        self.embeddings = EmbeddingsWithTokenTypes(
            vocab_size=vocab_size, token_types=token_types, embedding_size=dim, pad_token_id=pad_token_id,
            max_position_embeddings=max_position_embeddings
        )

        self.encoder = TransformerEncoder(
            n_heads=n_heads, embedding_size=dim, ffn_size=ffn_size, n_layers=n_layers)

    def forward(self, input_ids: Tensor, attention_mask: Tensor, token_type_ids: Optional[Tensor]) -> Tensor:
        hidden_states = self.embeddings(input_ids=input_ids, token_type_ids=token_type_ids)
        return self.encoder(hidden_states, mask=attention_mask.to(torch.bool))


class ElectraClassificationHead(Module):
    """Head for sentence-level classification tasks."""

    def __init__(self, dim: int, num_labels: int):
        super().__init__()
        self.dense = Linear(dim, dim)
        self.out_proj = Linear(dim, num_labels)

    def forward(self, features):
        x = features[:, 0]
        x = self.dense(x)
        x = gelu(x)
        x = self.out_proj(x)
        return x


class ElectraForSequenceClassification(Module):
    def __init__(
            self, vocab_size: int, token_types: int, dim: int, pad_token_id: int,
            max_position_embeddings: int, ffn_size: int, n_heads: int, n_layers: int,
            num_labels: int, activation: str
    ):
        super().__init__()

        self.num_labels = num_labels
        self.transformer = ElectraModel(
            vocab_size=vocab_size, token_types=token_types, dim=dim, pad_token_id=pad_token_id,
            max_position_embeddings=max_position_embeddings, ffn_size=ffn_size, n_heads=n_heads, n_layers=n_layers)
        self.classifier = ElectraClassificationHead(dim, num_labels)
        self.sigmoid = True if activation == 'sigmoid' else False

    def forward(self, input_ids: Tensor, attention_mask: Tensor, token_type_ids: Optional[Tensor] = None):
        tensor = self.transformer(input_ids, attention_mask, token_type_ids)
        tensor = self.classifier(tensor)
        return tensor.sigmoid() if self.sigmoid else tensor.softmax(dim=-1)

    @classmethod
    def from_pretrained(cls, model_path: str, num_labels: int, activation: str) -> 'ElectraForSequenceClassification':
        model = cls(
            vocab_size=30522, token_types=2, dim=768, pad_token_id=0, max_position_embeddings=512,
            ffn_size=3072, n_heads=12, n_layers=12, num_labels=num_labels, activation=activation
        )
        model.load_state_dict(state_dict=torch.load(os.path.join(model_path, 'pytorch_model.bin'), map_location='cpu'))
        model.eval()
        for p in model.parameters():
            p.requires_grad = False
        return model.half()

    def load_state_dict(self, state_dict, strict=True):
        del state_dict['electra.embeddings.position_ids']
        module_names = {'layers': 'layer', 'transformer': 'electra'}
        for key, value in list(state_dict.items()):
            new_key = key
            for name, old_name in module_names.items():
                new_key = new_key.replace(f'{old_name}.', f'{name}.')
            state_dict[new_key] = state_dict.pop(key)
        old_names = {
            'attention.out_lin': 'attention.output.dense',
            'norm1': 'attention.output.LayerNorm',
            'norm2': 'output.LayerNorm',
            'ffn.lin1': 'intermediate.dense',
            'ffn.lin2': 'output.dense'
        }
        for layer_ind in range(len(self.transformer.encoder.layers)):
            prefix = f'transformer.encoder.layers.{layer_ind}.'
            for p in ('weight', 'bias'):
                state_dict[f'{prefix}attention.pre_attention.{p}'] = transposed_cat([
                    state_dict.pop(f'{prefix}attention.self.{k}.{p}') for k in ('query', 'key', 'value')
                ])
                for new_name, old_name in old_names.items():
                    state_dict[f'{prefix}{new_name}.{p}'] = state_dict.pop(f'{prefix}{old_name}.{p}')
        return super().load_state_dict(state_dict, strict)
